/** 
@file unDeviceTypeCreator.ctl
  Functions used to create a new package from a given template, and if necessary from existing device type information (such as specific CTRL code)


@author Cyril Caillaba (EN-ICE-SCD)
@version 1.0.0
@date October 2012
*/

// Functions to manipulate front-end types
#uses "libunCore/unDeviceTypeCreatorFrontEndFunctions.ctl"
// Functions to manipulate device types
#uses "libunCore/unDeviceTypeCreatorDeviceFunctions.ctl"
// Functions to manipulate widget types
#uses "libunCore/unDeviceTypeCreatorWidgetFunctions.ctl"

// XML functions
#uses "CtrlXml"

//{@

//main
/**
  @par Description:
  Automatically creates a new package with default parameters.
  
  @par Usage: 
  Public
*/
main()
{  
  dyn_string dsExceptions;
  int iResult = unDeviceTypeCreator_createNewPackage(dsExceptions);  
  if (0 != iResult)
	{
		DebugN(dsExceptions);
	}
}

//unDeviceTypeCreator_createNewPackage
/**
  @par Description:
  Starts creating the package with the given information.
  
  @par Usage: 
  Public
  
	@param dsExceptions output, list of errors that occurred during the process
	@param bUiActive input, indicates if the function is called from the dedicated panel or not (default: not)
  @param sLoadFrom input string, file from which to load the structure of the package. If no file provided, the default one is loaded. If the default one doesn't exist, an exception is raised.
  
  @return Status code
     0 : No error
    -1 : Invalid config file
    -2 : Invalid parameters
    -3 : Invalid data configuration
    -4 : Error in generation
*/
int unDeviceTypeCreator_createNewPackage(dyn_string &dsExceptions, bool bUiActive = false, string sLoadFrom = "")
{
  g_bUiActive = bUiActive;
  
	// If the function is called from the panel, initialize the progress bar.
  if (g_bUiActive)
  {
    openProgressBar(
      "UNICOS package generation", 
      "copy.gif", 
      "", 
      "", 
      "", 
      2);
  }
  
  // -------------------------
  // ----- 1) Load parameters  
  // -------------------------
  
  // If not called from UI, the main parameters are not defined.
  // It is then necessary to load them from a file.
  if (!g_bUiActive)
  {
    
    int iRes = _unDeviceTypeCreator_loadParameters(sLoadFrom, dsExceptions);
    
    if ((dynlen(dsExceptions) > 0) || (0 != iRes))
    {
    	if (g_bUiActive)
      {
    		closeProgressBar();
      }
      return -1;
    }
  }


  
  // -------------------------
  // ----- 2) Check parameters
  // -------------------------
  
  int iRes = _unDeviceTypeCreator_checkParameters(dsExceptions);
  if ((dynlen(dsExceptions) > 0) || (0 != iRes))
  {
  	if (g_bUiActive)
    {
  		closeProgressBar();
    }
    return -2;
  }   
  
  
  
  // -------------------------
  // ----- 3) Check data
  // -------------------------
  
  int iRes = _unDeviceTypeCreator_checkData(dsExceptions);
  if ((dynlen(dsExceptions) > 0) || (0 != iRes))
  {
  	if (g_bUiActive)
    {
  		closeProgressBar();
    }
    return -3;
  }   
  
  
  
  
  // -------------------------
  // ----- 4) Init variables
  // -------------------------
  
  g_xmlPackageDescription = xmlNewDocument();  
  dynClear(g_dsLibList);
  dynClear(g_dsLibListUiOnly);
  
  
  // -------------------------
  // ----- 5) Generate package
  // -------------------------
  int iRes = _unDeviceTypeCreator_generatePackage(dsExceptions);
  
  if ((dynlen(dsExceptions) > 0) || (0 != iRes))
  {
  	if (g_bUiActive)
    {
  		closeProgressBar();
    }
    return -4;
  }  
	
	// Close progress bar once process over
	if (g_bUiActive)
  {
		closeProgressBar();
  }
  
  return 0;
}


// ----------------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------------
// ------------------------------------ GLOBAL VARIABLES USED THROUGHOUT THE PROCESS ------------------------------------
// ----------------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------------
// Note: all variables have to be declared as global.
// A non global variable is lost between two calls from different sections of a panel (e.g. initialize of the panel and click of a button).
// ----------------------------------------------------------------------------------------------------------------------



// Is the function called from UI or not? If not, nothing related to widgets must be shown (such as exception display or progress bar)
global bool g_bUiActive = false;

// Parameters
// Name of the new package
global string g_sPackageName;
// Prefix used in this package
global string g_sPackagePrefix;
// Path to read the template from
global string g_sTemplatePath;
// Path to write the new package to
global string g_sNewPackagePath;

// Mapping of device types to front-end types
global mapping g_mFrontEndToDeviceMapping;
// Mapping of widget types to device types
global mapping g_mDeviceToWidgetMapping;

// List of front-end types
// Each element contains both the name of the type and wether it already exists or not
global dyn_dyn_anytype g_ddaFrontEndTypeList;
// List of device types
// Each element contains both the name of the type and wether it already exists or not
global dyn_dyn_anytype g_ddaDeviceTypeList;
// List of widget types
// Each element contains both the name of the type and wether it already exists or not
global dyn_dyn_anytype g_ddaWidgetTypeList;

// The version of the generated package
// By default 0.0.1, but if a package is being edited its current version is used
global string g_sNewPackageVersion = "0.0.1";
    
// Location of the loaded package files
global string mLoadedPackageLocation;
    
// List of files used by the loaded package
global dyn_string dsLoadedPackageFiles;

// Are we editing an existring package or creating a new one?
global bool bUpdatingExistingPackage = false;

// Mapping of device type information
// Key: device type (string)
// Value: mapping
// ------ Key: scripts
// ------ Value: mapping
// ------------- Key: script Id (ex: TAG_DT_SCRIPT_setConfig)
// ------------- Value: script content (CTRL code)
global mapping g_mDeviceTypeInfo;

// List of libs to be loaded with the UI and CTRL managers
global dyn_string g_dsLibList;
// List of libs to be loaded by only the UI manager
global dyn_string g_dsLibListUiOnly;

// Content of the template description XML file
global unsigned g_xmlTemplateFileList;
// First Node of this file
global unsigned g_xmlTemplateFileListFirstChild;
// XML file to describe the component (used to install package)
global unsigned g_xmlPackageDescription;
// First node of this file
global unsigned g_xmlPackageDescriptionFirstChild;

// File from which the package structure has been loaded
global string sLoadPackacteStructureFrom;



// ---------------------------------------------------------------------------------------------------------------
// ---------------------------------------------------------------------------------------------------------------
// ------------------------------------ CONSTANTS USED THROUGHOUT THE PROCESS ------------------------------------
// ---------------------------------------------------------------------------------------------------------------
// ---------------------------------------------------------------------------------------------------------------


// Each element of the list of front-end, device and widget types contains two elements
// The first is the name of this type, the second indicates if it exists or not
const int LIST_NAME_INDEX = 1;
const int LIST_EXISTS_INDEX = 2;

// Each type of type (... i.e. front-end, device, widget) is represented by a int. This allows to use generic functions instead of one for front-ends, one for devices and one for widgets.
const int CURRENT_TYPE_FRONT_END = 3;
const int CURRENT_TYPE_DEVICE    = 2;
const int CURRENT_TYPE_WIDGET    = 1;
const int CURRENT_TYPE_NONE      = 0;

// Doxygen tag indicating the beginning of the file content
const string DOXYGEN_START_TAG = "//@{";
// Doxygen tag indicating the end of the file content
const string DOXYGEN_END_TAG = "//@}";

// Content to replace in files
const string TAG_DeviceType = "TAG_DeviceType";
const string TAG_DEVICETYPE = "TAG_DEVICETYPE";

const string TAG_FrontEndType = "TAG_FrontEndType";
const string TAG_FRONTENDTYPE = "TAG_FRONTENDTYPE";

const string TAG_WidgetType = "TAG_WidgetType"; 
const string TAG_WIDGETTYPE = "TAG_WIDGETTYPE";

const string TAG_package = "TAG_package";
const string TAG_PACKAGE = "TAG_PACKAGE";

const string TAG_prefix = "TAG_prefix";
const string TAG_PREFIX = "TAG_PREFIX";

// Folder to look for device type info
const string DEVICE_TYPE_INFO_FOLDER = "/deviceTypeInfo/";

// Default device type parameters file
const string DEFAULT_DEVICE_TYPE_XML = "_default.xml";

// Default file to load parameters from
const string DEFAULT_PARAMETERS_FILE = "unDeviceTypeCreatorFields.temp";

// Hard-coded variable of the last known validated package template.
// If the read template version is different, we raise an error and stop execution.
const string REQUIRED_TEMPLATE_VERSION = "2.0.0";


// Constant strings used in XML document generation (nodes name)
const string XML_NODE_COMPONENT                      = "component";
const string XML_NODE_NAME                           = "name";
const string XML_NODE_VERSION                        = "version";
const string XML_NODE_REQUIRED                       = "required";
const string XML_NODE_REQUIREDPVSS                   = "required_pvss_version";
const string XML_NODE_REQUIREDINSTALLER              = "required_installer_version";
const string XML_NODE_DATE                           = "date";
const string XML_NODE_CONFIG                         = "config";
const string XML_NODE_PREINIT                        = "preinit";
const string XML_NODE_INIT                           = "init";
const string XML_NODE_POSTINSTALL                    = "postInstall";
const string XML_NODE_DPLIST                         = "dplist";
const string XML_NODE_ATTRIBUTE_WIDGET_LIST_TEMPLATE = "widgetListTemplate";
const string XML_NODE_FILE                           = "file";
const string XML_NODE_SCRIPT                         = "script";
const string XML_NODE_PANEL                          = "panel";
const string XML_NODE_IMAGE                          = "image";
const string XML_NODE_ATTRIBUTE_CONCERNS             = "concerns";
const string XML_NODE_ATTRIBUTE_SUBTYPE              = "subType";
const string XML_NODE_ATTRIBUTE_UIONLY               = "configUiOnly";
const string XML_NODE_ATTRIBUTE_CASE                 = "particularCase";
const string XML_NODE_ATTRIBUTE_VERSION              = "templateVersion";

// Strings used as XML tags or attributes in the CTRL source template files
const string XML_TEMPLATE_CTRLSCRIPT                   = "ctrlScript";
const string XML_TEMPLATE_CTRLSCRIPT_ATTR_FILENAME     = "fileName";
const string XML_TEMPLATE_CTRLSCRIPT_ATTR_SUBTYPES     = "browseSubTypes";
const string XML_TEMPLATE_CTRLSCRIPT_ATTR_SUPERTYPES   = "browseSuperTypes";
const string XML_TEMPLATE_CTRLSCRIPT_ATTR_NEW_FRONTEND = "concernsNewFrontEndType";
const string XML_TEMPLATE_CTRLSCRIPT_ATTR_DUPL_FE_DEV  = "concernsFrontEndAndDevice";
const string XML_TEMPLATE_CTRLSCRIPT_ATTR_NEW_DEVICE   = "concernsNewDeviceType";
const string XML_TEMPLATE_CTRLSCRIPT_ATTR_DUPL_DEV_W   = "concernsDeviceAndWidget";
const string XML_TEMPLATE_CTRLSCRIPT_ATTR_NEW_WIDGET   = "concernsNewWidgetType";
const string XML_TEMPLATE_CTRLSCRIPT_HEADER            = "header";
const string XML_TEMPLATE_CTRLSCRIPT_ELEMENT_LIST      = "elementList";
const string XML_TEMPLATE_FUNCTION                     = "function";
const string XML_TEMPLATE_FUNCTION_ATTR_NAME           = "name";
const string XML_TEMPLATE_FUNCTION_HEADER              = "header";
const string XML_TEMPLATE_FUNCTION_CONTENT             = "content";
const string XML_TEMPLATE_CONSTANT                     = "constant";
const string XML_TEMPLATE_CONSTANT_ATTR_NOREPLACE      = "noReplace";
const string XML_TEMPLATE_CONSTANT_ATTR_NAME           = "name";
const string XML_TEMPLATE_CONSTANT_COMMENT             = "comment";
const string XML_TEMPLATE_ATTR_DUPLICATE_COMMENT       = "duplicateComment";
const string XML_TEMPLATE_CONSTANT_CONTENT             = "content";
const string XML_TEMPLATE_ATTR_DUPLICATE_ELEMENT       = "duplicate";

// Constants used for import and export
const string XML_IMPORT_EXPORT_NODE_PACKAGE         = "package";
const string XML_IMPORT_EXPORT_NODE_LISTS           = "lists";
const string XML_IMPORT_EXPORT_NODE_FRONTENDLIST    = "frontEndList";
const string XML_IMPORT_EXPORT_NODE_DEVICELIST      = "deviceList";
const string XML_IMPORT_EXPORT_NODE_WIDGETLIST      = "widgetList";
const string XML_IMPORT_EXPORT_NODE_MAPPING         = "mapping";
const string XML_IMPORT_EXPORT_NODE_FRONTENDDEVICE  = "frontEndDevice";
const string XML_IMPORT_EXPORT_NODE_DEVICEWIDGET    = "deviceWidget";
const string XML_IMPORT_EXPORT_NODE_FRONTENDTYPE    = "frontEndType";
const string XML_IMPORT_EXPORT_NODE_DEVICETYPE      = "deviceType";
const string XML_IMPORT_EXPORT_NODE_WIDGETTYPE      = "widgetType";
const string XML_IMPORT_EXPORT_ATTR_NAME            = "name";
const string XML_IMPORT_EXPORT_ATTR_PREFIX          = "prefix";
const string XML_IMPORT_EXPORT_ATTR_EXISTS          = "exists";
	
// Device type information indexes
const string DEVICE_TYPE_INDEX_DPLIST = "dplist";
const string DEVICE_TYPE_INDEX_SCRIPT = "scripts";
const string DEVICE_TYPE_INDEX_IGNORE_FUNCTIONS = "ignoreFunctions";
const string DEVICE_TYPE_INDEX_IGNORE_PANELS = "ignorePanels";
const string DEVICE_TYPE_INDEX_IGNORE_SCRIPTS = "ignoreScripts";
const string DEVICE_TYPE_INDEX_IGNORE_DPLISTS = "ignoreDpLists";
const string DEVICE_TYPE_INDEX_IGNORE_IMAGES = "ignoreImages";
const string DEVICE_TYPE_INDEX_INCLUDE = "include";
// Beginning of a custom script tag
const string DEVICE_TYPE_CUSTOM_SCRIPT_TAG_START = "TAG_SCRIPT_DEVICE_TYPE";

const string EMPTY_WIDGET_IDENTIFIER = "--empty--";
// --------------------------------------------------------------------------------------------
// --------------------------------------------------------------------------------------------
// ------------------------------------ INTERNAL FUNCTIONS ------------------------------------
// --------------------------------------------------------------------------------------------
// --------------------------------------------------------------------------------------------

//_unDeviceTypeCreator_loadParameters
/**
  @par Description:
  Load parameters from a file.
  
  @par Usage: 
  Internal.
  
  @param sLoadFrom     input, file from which to load the structure of the package.
	@param dsExceptions output, list of errors that occurred during the process
  
  @return -1 if an error occured, 0 otherwise
*/
int _unDeviceTypeCreator_loadParameters(string sLoadFrom, dyn_string &dsExceptions)
{
	string sFileName = sLoadFrom;
	
	if (("" == sFileName) || (!isfile(sFileName)))
	{
		sFileName = getPath(DATA_REL_PATH) + DEFAULT_PARAMETERS_FILE;
	}
	
	
  string sFileContent;
	if (isfile(sFileName))
  {
    fileToString(sFileName, sFileContent);
  }
	else
	{
		return -1;
	}
	
	dyn_string dsFileContent = strsplit(sFileContent, "\n");
	
	string sTempTemplatePath = dsFileContent[1];
  string sTempNewPackagePath;
  string sTempStructureFilePath;
  
  if (dynlen(dsFileContent) > 1)
  {
    sTempNewPackagePath = dsFileContent[2];  
  }
  
  if (dynlen(dsFileContent) > 2)
  {
    sTempStructureFilePath = dsFileContent[3];
  }
	
	_unDeviceTypeCreator_importStructure(sTempStructureFilePath, dsExceptions);
	if (dynlen(dsExceptions) > 0)
	{
		return -1;
	}
	
	g_sTemplatePath = sTempTemplatePath;
	g_sNewPackagePath = sTempNewPackagePath;
	
	
	return 0;
}

//_unDeviceTypeCreator_checkParameters
/**
  @par Description:
  Check if the parameters (paths, name, prefix) are valid before proceeding to generation.
  
  @par Usage: 
  Internal
  
	@param dsExceptions output, list of errors that occurred during the process
  
  @return -1 if an error occured, 0 otherwise
*/
int _unDeviceTypeCreator_checkParameters(dyn_string &dsExceptions)
{
	// --------------------------------------------------------------------------------
	// ---------------------------------------- 1) Check template path
	// --------------------------------------------------------------------------------
	
	if (g_sTemplatePath == "")
  {
    string sError = getCatStr("unDeviceTypeCreator","GENERATE-EMPTYTEMPLATELOCATION");      
    fwException_raise(
        dsExceptions,
        "ERROR",
        sError,
        "GENERATE-EMPTYTEMPLATELOCATION"
    ); 
    return -1;     
  }
	
	
	// Template is supposed to be valid.
  // So the only verification needed is the template version.
  // To avoid errors we need to check data until the version verification (i.e file readable, syntax errors)  
  
  // Check if the folder contains the template by checking if templateContent.xml exists and is readable
  if ((access(g_sTemplatePath + "/templateContent.xml",F_OK) != 0) || (access(g_sTemplatePath + "/templateContent.xml",R_OK) != 0))
  {
    string sError = getCatStr("unDeviceTypeCreator","GENERATE-INVALIDTEMPLATELOCATION");
    strreplace(sError,"TEMPLATESOURCEPATH", g_sTemplatePath);
      
    fwException_raise(
        dsExceptions,
        "ERROR",
        sError,
        "GENERATE-INVALIDTEMPLATELOCATION"
    ); 
    return -1; 
  }     
	
	string errMsg;
  int errLine;
  int errColumn;
  g_xmlTemplateFileList = xmlDocumentFromFile(
      g_sTemplatePath + "/templateContent.xml",
      errMsg,
      errLine,
      errColumn
      );
  
  // XML syntax error ?
  if (g_xmlTemplateFileList == -1)
  {
    string sError = getCatStr("unDeviceTypeCreator","GENERATE-XMLSYNTAXERROR");
    strreplace(sError,"ERRMSG",errMsg);
    strreplace(sError,"LINENUM",errLine);
    strreplace(sError,"COLUMNUM",errColumn);
      
    fwException_raise(
        dsExceptions,
        "ERROR",
        sError,
        "GENERATE-XMLSYNTAXERROR"
    ); 
    return -1; 
  }
  
  // Access "component" tag
  g_xmlTemplateFileListFirstChild = xmlFirstChild(g_xmlTemplateFileList);
  
  // Read version
  string sTemplateVersion;
  // Template version MUST be the one known
  xmlGetElementAttribute(g_xmlTemplateFileList, g_xmlTemplateFileListFirstChild, XML_NODE_ATTRIBUTE_VERSION, sTemplateVersion);
  if (sTemplateVersion != REQUIRED_TEMPLATE_VERSION)
  {
    string sError = getCatStr("unDeviceTypeCreator","GENERATE-INVALIDTEMPLATEVERSION");
    strreplace(sError,"VERSION",sTemplateVersion);
    strreplace(sError,"EXPECTED",REQUIRED_TEMPLATE_VERSION);
      
    fwException_raise(
        dsExceptions,
        "ERROR",
        sError,
        "GENERATE-INVALIDTEMPLATEVERSION"
    ); 
    return -1;   
  }
	
	
	
	// --------------------------------------------------------------------------------
	// ---------------------------------------- 2) Check package to create path
	// --------------------------------------------------------------------------------
	
	if (g_sNewPackagePath == "")
  {
    string sError = getCatStr("unDeviceTypeCreator","GENERATE-EMPTYDESTINATIONPATH");
      
    fwException_raise(
        dsExceptions,
        "ERROR",
        sError,
        "GENERATE-EMPTYDESTINATIONPATH"
    ); 
    return -1; 
  }
	
	
	
	// --------------------------------------------------------------------------------
	// ---------------------------------------- 3) Check prefix
	// --------------------------------------------------------------------------------
	
  if (g_sPackagePrefix == "")
  {
    string sError = getCatStr("unDeviceTypeCreator","GENERATE-EMPTYPREFIX");
      
    fwException_raise(
        dsExceptions,
        "ERROR",
        sError,
        "GENERATE-EMPTYPREFIX"
    ); 
    return -1; 
  }
	
	
	
	// --------------------------------------------------------------------------------
	// ---------------------------------------- 4) Check package name
	// --------------------------------------------------------------------------------
	  if (g_sPackageName == "")
  {
    string sError = getCatStr("unDeviceTypeCreator","GENERATE-EMPTYPACKAGENAME");
      
    fwException_raise(
        dsExceptions,
        "ERROR",
        sError,
        "GENERATE-EMPTYPACKAGENAME"
    ); 
    return -1;   
  }
	
	
	return 0;
}

//_unDeviceTypeCreator_checkData
/**
  @par Description:
  Check if the data (links between front-end, device and widget types) is valid before proceeding to generation.
  
  @par Usage: 
  Internal
  
	@param dsExceptions output, list of errors that occurred during the process
  
  @return -1 if an error occured, 0 otherwise
*/
int _unDeviceTypeCreator_checkData(dyn_string &dsExceptions)
{
  // The user can do anything with the data, but some combinations don't make sense and we need to prevent them.
	
  // 1. A device type must always have at least one widget type associated.
  for (int i = 1 ; i <= mappinglen(g_mDeviceToWidgetMapping) ; i++)
  {
    // No mapped widget
    if (dynlen(mappingGetValue(g_mDeviceToWidgetMapping,i)) == 0)
    {
      string sError = getCatStr("unDeviceTypeCreator","GENERATE-DEVICEHASNOWIDGET");
      strreplace(sError,"DEVICE",mappingGetKey(g_mDeviceToWidgetMapping,i));
      
      fwException_raise(
          dsExceptions,
          "ERROR",
          sError,
          "GENERATE-DEVICEHASNOWIDGET"
      );
      
      return -1;
    }  
  }  
  	
  // 2. If a widget already exists, then its associated device must also already exist (consequence of check #4)
  for (int i = 1 ; i <= mappinglen(g_mDeviceToWidgetMapping) ; i++)
  {
    bool bDeviceExists = _unDeviceTypeCreator_typeExists(mappingGetKey(g_mDeviceToWidgetMapping,i),CURRENT_TYPE_DEVICE,dsExceptions);
    if (dynlen(dsExceptions) > 0)
    {
      return -1;
    }
    
    for (int j = 1 ; j <= dynlen(mappingGetValue(g_mDeviceToWidgetMapping,i)) ; j++)
    {
      bool bWidgetExists = _unDeviceTypeCreator_typeExists(mappingGetValue(g_mDeviceToWidgetMapping,i)[j],CURRENT_TYPE_WIDGET,dsExceptions);
      if (dynlen(dsExceptions) > 0)
      {
        return -1;
      }
      if (bWidgetExists && !bDeviceExists)
      {
        string sError = getCatStr("unDeviceTypeCreator","GENERATE-WIDGETEXISTSDEVICENOT");
        strreplace(sError,"DEVICE",mappingGetKey(g_mDeviceToWidgetMapping,i));
        strreplace(sError,"WIDGET",mappingGetValue(g_mDeviceToWidgetMapping,i)[j]);
        fwException_raise(
            dsExceptions,
            "ERROR",
            sError,
            "GENERATE-WIDGETEXISTSDEVICENOT"
        );
        return -1;
      }
    }  
  } 
 		
  // 3. A device type must always be mapped to at least one front-end type. 
  for (int i = 1 ; i <= mappinglen(g_mDeviceToWidgetMapping) ; i++)
  {
    string currentDevice = mappingGetKey(g_mDeviceToWidgetMapping,i);
    bool currentDeviceIsContained = false;
    for (int j = 1 ; j <= mappinglen(g_mFrontEndToDeviceMapping) ; j++)
    {
      if (dynContains(mappingGetValue(g_mFrontEndToDeviceMapping,j),currentDevice) > 0)
      {
        currentDeviceIsContained = true;
      }                      
    }  
    if (!currentDeviceIsContained)
    {
      string sError = getCatStr("unDeviceTypeCreator","GENERATE-DEVICEHASNOFRONTEND");
      strreplace(sError,"DEVICE",currentDevice);
      
      fwException_raise(
          dsExceptions,
          "ERROR",
          sError,
          "GENERATE-DEVICEHASNOFRONTEND"
      );
      return -1;
    }    
  }
	
	return 0;
}

//_unDeviceTypeCreator_typeExists
/**
  @par Description:
  Check if given type (front-end, device or widget) is flagged as already existing.
  
  @par Usage: 
  Internal
  
	@param sTypeName     input, the name of the type to check.
	@param iTypeType     input, indicates wether it is a front-end (CURRENT_TYPE_FRONT_END), a device (CURRENT_TYPE_DEVICE), a widget type (CURRENT_TYPE_WIDGET) or none (CURRENT_TYPE_NONE, results in an error).
	@param dsExceptions output, list of errors that occurred during the process.
  
  @return true if the type exists, false if an error occurs.
*/
bool _unDeviceTypeCreator_typeExists(string sTypeName, int iTypeType, dyn_string &dsExceptions)
{
  switch (iTypeType)
  {
    case CURRENT_TYPE_DEVICE:
    {  
      for (int i = 1 ; i <= dynlen(g_ddaDeviceTypeList) ; i++)
      {
        if (g_ddaDeviceTypeList[i][LIST_NAME_INDEX] == sTypeName)
        {
          return g_ddaDeviceTypeList[i][LIST_EXISTS_INDEX];
        }
      }
      
      break;
    }
		
    case CURRENT_TYPE_FRONT_END:
    {  
      for (int i = 1 ; i <= dynlen(g_ddaFrontEndTypeList) ; i++)
      {
        if (g_ddaFrontEndTypeList[i][LIST_NAME_INDEX] == sTypeName)
        {
          return g_ddaFrontEndTypeList[i][LIST_EXISTS_INDEX];
        }
      }
      
      break;
    }
		
    case CURRENT_TYPE_WIDGET:
    {  
      for (int i = 1 ; i <= dynlen(g_ddaWidgetTypeList) ; i++)
      {
        if (g_ddaWidgetTypeList[i][LIST_NAME_INDEX] == sTypeName)
        {
          return g_ddaWidgetTypeList[i][LIST_EXISTS_INDEX];
        }
      }
      
      break;
    }
    default: 
      break;
  
  }
  
  // Error, the type was found in none of the original lists
  string sError = getCatStr("unDeviceTypeCreator","GENERATE-UNKNOWNTYPE");
  strreplace(sError,"TYPE",sTypeName);
      
  fwException_raise(
      dsExceptions,
      "ERROR",
      sError,
      "GENERATE-UNKNOWNTYPE"
  );
  return false;
  
}

//_unDeviceTypeCreator_loadDeviceTypeInfo
/**
  @par Description:
  Load information (custom code or dplist configuration) for device types. Any information found is loaded, even if the device is not concerned by the package to be generated.
	All this information is saved in the mapping g_mDeviceTypeInfo.
  
  @par Usage: 
  Internal
  
	@param dsExceptions output, list of errors that occurred during the process.
*/
void _unDeviceTypeCreator_loadDeviceTypeInfo(dyn_string &dsExceptions)
{

	// Constant strings used in the device type information XML file
	const string XML_NODE_DEVICE_ROOT               = "deviceType";
	const string XML_NODE_DEVICE_SCRIPT             = "script";
	const string XML_NODE_DEVICE_SCRIPT_ATTR_NAME   = "name";
	const string XML_NODE_DEVICE_DPLIST             = "dplist";
	const string XML_NODE_DEVICE_DPLIST_ATTR_NAME   = "name";
	const string XML_NODE_DEVICE_IGNORE					    = "ignore";
	const string XML_NODE_DEVICE_IGNORE_ATTR_SCRIPT = "script";
	const string XML_NODE_DEVICE_IGNORE_ATTR_PANEL  = "panel";
	const string XML_NODE_DEVICE_IGNORE_ATTR_IMAGE  = "image";
	const string XML_NODE_DEVICE_IGNORE_ATTR_DPLIST = "dplist";
	const string XML_NODE_DEVICE_INCLUDE_ATTR       = "include";

	mappingClear(g_mDeviceTypeInfo);
	
  for (int i = 1 ; i <= dynlen(g_ddaDeviceTypeList) ; i++)
  {	
		string sDeviceType = g_ddaDeviceTypeList[i][LIST_NAME_INDEX];
		string sFileName = g_sTemplatePath;
		if (sFileName[strlen(sFileName)] != "/")
		{
			sFileName += "/";
		}
		
		sFileName += DEVICE_TYPE_INFO_FOLDER + sDeviceType + ".xml";
		
		if (access(sFileName, R_OK) == 0)
		{
			string sErrMsg;
			int iErrLine;
			int iErrColumn;
			unsigned xmlDeviceTypeInfo = xmlDocumentFromFile(sFileName, sErrMsg, iErrLine, iErrColumn);
			
			// Syntax error?
			if (-1 == xmlDeviceTypeInfo)
			{
				string sError = "Error in file \"" + sFileName + "\"\n" + getCatStr("unDeviceTypeCreator","LOAD-XMLSYNTAXERROR");
				strreplace(sError,"ERRMSG",sErrMsg);
				strreplace(sError,"LINENUM",iErrLine);
				strreplace(sError,"COLUMNUM",iErrColumn);
				
				fwException_raise(
						dsExceptions,
						"ERROR",
						sError,
						"LOAD-XMLSYNTAXERROR"
				); 
				return; 
			}
			
			unsigned xmlNodeRoot = xmlFirstChild(xmlDeviceTypeInfo);
		 
			if (xmlNodeName(xmlDeviceTypeInfo, xmlNodeRoot) != XML_NODE_DEVICE_ROOT)
			{
				string sError = "Error in file \"" + sFileName + "\"\n" + getCatStr("unDeviceTypeCreator","LOAD-INCORRECTXMLSTRUCTURE");
				strreplace(sError,"TAG",xmlNodeName(xmlDeviceTypeInfo, xmlNodeRoot));

				fwException_raise(
						dsExceptions,
						"ERROR",
						sError,
						"LOAD-INCORRECTXMLSTRUCTURE"
					); 
				return; 
			}
			
			unsigned xmlNodeBrowse = xmlFirstChild(xmlDeviceTypeInfo, xmlNodeRoot);
			mapping mScripts;
			mapping mDpLists;
			dyn_string dsIgnoreScripts = makeDynString();
			dyn_string dsIgnorePanels = makeDynString();
			dyn_string dsIgnoreDpLists = makeDynString();
			dyn_string dsIgnoreImages = makeDynString();
			dyn_string dsInclude = makeDynString();
			while (-1 != xmlNodeBrowse)
			{
				switch (xmlNodeName(xmlDeviceTypeInfo, xmlNodeBrowse))
				{
					case XML_NODE_DEVICE_SCRIPT:
					{
						mapping mAttributes = xmlElementAttributes(xmlDeviceTypeInfo, xmlNodeBrowse);
						
						string sScriptId;
						
						if (mappingHasKey(mAttributes, XML_NODE_DEVICE_SCRIPT_ATTR_NAME))
						{
							sScriptId = mAttributes[XML_NODE_DEVICE_SCRIPT_ATTR_NAME];
						}
						
						if ("" == sScriptId)
						{
							string sError = "Error in file \"" + sFileName + "\"\n" + getCatStr("unDeviceTypeCreator","LOAD-INCORRECTXMLSTRUCTURE2");

							fwException_raise(
									dsExceptions,
									"ERROR",
									sError,
									"LOAD-INCORRECTXMLSTRUCTURE"
								);
							return; 
						}
						
						string sScriptContent = xmlNodeValue(xmlDeviceTypeInfo, xmlFirstChild(xmlDeviceTypeInfo,xmlNodeBrowse));            
            strreplace(sScriptContent, "\r", "");
            sScriptContent = strrtrim(sScriptContent);
            sScriptContent = strltrim(sScriptContent, "\n");
						mScripts[sScriptId] = "//begin_" + sScriptId + "\n" + sScriptContent + "\n" + "//end_" + sScriptId;
						break;
					}
					
					
					case XML_NODE_DEVICE_DPLIST:
					{
						mapping mAttributes = xmlElementAttributes(xmlDeviceTypeInfo, xmlNodeBrowse);
						
						string sDplId;
						
						if (mappingHasKey(mAttributes, XML_NODE_DEVICE_DPLIST_ATTR_NAME))
						{
							sDplId = mAttributes[XML_NODE_DEVICE_DPLIST_ATTR_NAME];
						}
						
						if ("" == sDplId)
						{
							string sError = "Error in file \"" + sFileName + "\"\n" + getCatStr("unDeviceTypeCreator","LOAD-INCORRECTXMLSTRUCTURE2");

							fwException_raise(
									dsExceptions,
									"ERROR",
									sError,
									"LOAD-INCORRECTXMLSTRUCTURE"
								);
							return; 
						}          
						
						mDpLists[sDplId] = xmlNodeValue(xmlDeviceTypeInfo, xmlFirstChild(xmlDeviceTypeInfo,xmlNodeBrowse));
						break;
					}
					
					case XML_NODE_DEVICE_IGNORE:
					{
						mapping mAttributes = xmlElementAttributes(xmlDeviceTypeInfo, xmlNodeBrowse);
						
						if (mappingHasKey(mAttributes, XML_NODE_DEVICE_IGNORE_ATTR_SCRIPT))
						{
							string sIgnoreScript = mAttributes[XML_NODE_DEVICE_IGNORE_ATTR_SCRIPT];
							dynAppend(dsIgnoreScripts, sIgnoreScript);
						}
						else if (mappingHasKey(mAttributes, XML_NODE_DEVICE_IGNORE_ATTR_PANEL))
						{
							string sIgnorePanel = mAttributes[XML_NODE_DEVICE_IGNORE_ATTR_PANEL];
							dynAppend(dsIgnorePanels, sIgnorePanel);
						}
						else if (mappingHasKey(mAttributes, XML_NODE_DEVICE_IGNORE_ATTR_IMAGE))
						{
							string sIgnoreImage = mAttributes[XML_NODE_DEVICE_IGNORE_ATTR_IMAGE];
							dynAppend(dsIgnoreImages, sIgnoreImage);
						}
						else if (mappingHasKey(mAttributes, XML_NODE_DEVICE_IGNORE_ATTR_DPLIST))
						{
							string sIgnoreDpl = mAttributes[XML_NODE_DEVICE_IGNORE_ATTR_DPLIST];
							dynAppend(dsIgnoreDpLists, sIgnoreDpl);
						}
						else 
						{
							string sError = "Error in file \"" + sFileName + "\"\n" + getCatStr("unDeviceTypeCreator","LOAD-INCORRECTXMLSTRUCTURE2");

							fwException_raise(
									dsExceptions,
									"ERROR",
									sError,
									"LOAD-INCORRECTXMLSTRUCTURE"
								);
							return; 
						}
						break;
					}

          case XML_NODE_DEVICE_INCLUDE_ATTR:
          {
						mapping attrs = xmlElementAttributes(xmlDeviceTypeInfo, xmlNodeBrowse);
						dynAppend(dsInclude, attrs["file"]);
						break;
          }
				}
				
				xmlNodeBrowse = xmlNextSibling(xmlDeviceTypeInfo, xmlNodeBrowse);
			}
			
			mapping mTemp;
			mTemp[DEVICE_TYPE_INDEX_SCRIPT] = mScripts;
			mTemp[DEVICE_TYPE_INDEX_DPLIST] = mDpLists;
			mTemp[DEVICE_TYPE_INDEX_IGNORE_PANELS] = dsIgnorePanels;
			mTemp[DEVICE_TYPE_INDEX_IGNORE_SCRIPTS] = dsIgnoreScripts;
			mTemp[DEVICE_TYPE_INDEX_IGNORE_DPLISTS] = dsIgnoreDpLists;
			mTemp[DEVICE_TYPE_INDEX_IGNORE_IMAGES] = dsIgnoreImages;
			mTemp[DEVICE_TYPE_INDEX_INCLUDE] = dsInclude;
			g_mDeviceTypeInfo[sDeviceType] = mTemp;
			
		}
		

    if (dynlen(dsExceptions) > 0)
    {
      return;
    }          
  }
	
	// For every other device, open the default file and read
	for (int i = 1 ; i <= dynlen(g_ddaDeviceTypeList) ; i++)
	{
		const string sDeviceType = g_ddaDeviceTypeList[i][LIST_NAME_INDEX];
		string sDefaultDeviceXml = g_sTemplatePath + DEVICE_TYPE_INFO_FOLDER + DEFAULT_DEVICE_TYPE_XML;
	
		if (access(sDefaultDeviceXml, R_OK) == 0)
		{
			string sErrMsg;
			int iErrLine;
			int iErrColumn;
			unsigned xmlDeviceTypeInfo = xmlDocumentFromFile(sDefaultDeviceXml, sErrMsg, iErrLine, iErrColumn);
			
			// Syntax error?
			if (-1 == xmlDeviceTypeInfo)
			{
				string sError = "Error in file \"" + sDefaultDeviceXml + "\"\n" + getCatStr("unDeviceTypeCreator","LOAD-XMLSYNTAXERROR");
				strreplace(sError,"ERRMSG",sErrMsg);
				strreplace(sError,"LINENUM",iErrLine);
				strreplace(sError,"COLUMNUM",iErrColumn);
				
				fwException_raise(
						dsExceptions,
						"ERROR",
						sError,
						"LOAD-XMLSYNTAXERROR"
				); 
				return; 
			}
			
			unsigned xmlNodeRoot = xmlFirstChild(xmlDeviceTypeInfo);
		 
			if (xmlNodeName(xmlDeviceTypeInfo, xmlNodeRoot) != XML_NODE_DEVICE_ROOT)
			{
				string sError = "Error in file \"" + sFileName + "\"\n" + getCatStr("unDeviceTypeCreator","LOAD-INCORRECTXMLSTRUCTURE");
				strreplace(sError,"TAG",xmlNodeName(xmlDeviceTypeInfo, xmlNodeRoot));

				fwException_raise(
						dsExceptions,
						"ERROR",
						sError,
						"LOAD-INCORRECTXMLSTRUCTURE"
					); 
				return; 
			}
			
			unsigned xmlNodeBrowse = xmlFirstChild(xmlDeviceTypeInfo, xmlNodeRoot);
			mapping mScripts;
			mapping mDpLists;
			dyn_string dsIgnorePanels = makeDynString();
			dyn_string dsIgnoreScripts = makeDynString();
			dyn_string dsIgnoreFunctions = makeDynString();
			dyn_string dsIgnoreDpLists = makeDynString();
			dyn_string dsIgnoreImages = makeDynString();
			dyn_string dsInclude = makeDynString();
			
			if (mappingHasKey(g_mDeviceTypeInfo, sDeviceType))
			{
				mScripts = g_mDeviceTypeInfo[sDeviceType][DEVICE_TYPE_INDEX_SCRIPT];
				mDpLists = g_mDeviceTypeInfo[sDeviceType][DEVICE_TYPE_INDEX_DPLIST];
				dsIgnorePanels = g_mDeviceTypeInfo[sDeviceType][DEVICE_TYPE_INDEX_IGNORE_PANELS];
				dsIgnoreScripts = g_mDeviceTypeInfo[sDeviceType][DEVICE_TYPE_INDEX_IGNORE_SCRIPTS];
				dsIgnoreDpLists = g_mDeviceTypeInfo[sDeviceType][DEVICE_TYPE_INDEX_IGNORE_DPLISTS];
				dsIgnoreImages = g_mDeviceTypeInfo[sDeviceType][DEVICE_TYPE_INDEX_IGNORE_IMAGES];
				dsInclude = g_mDeviceTypeInfo[sDeviceType][DEVICE_TYPE_INDEX_INCLUDE];
			}
			while (-1 != xmlNodeBrowse)
			{
				switch (xmlNodeName(xmlDeviceTypeInfo, xmlNodeBrowse))
				{
					case XML_NODE_DEVICE_SCRIPT:
					{
						mapping mAttributes = xmlElementAttributes(xmlDeviceTypeInfo, xmlNodeBrowse);
						
						string sScriptId;
						
						if (mappingHasKey(mAttributes, XML_NODE_DEVICE_SCRIPT_ATTR_NAME))
						{
							sScriptId = mAttributes[XML_NODE_DEVICE_SCRIPT_ATTR_NAME];
						}
						
						if ("" == sScriptId)
						{
							string sError = "Error in file \"" + sFileName + "\"\n" + getCatStr("unDeviceTypeCreator","LOAD-INCORRECTXMLSTRUCTURE2");

							fwException_raise(
									dsExceptions,
									"ERROR",
									sError,
									"LOAD-INCORRECTXMLSTRUCTURE"
								);
							return; 
						}
						
						string sScriptContent = xmlNodeValue(xmlDeviceTypeInfo, xmlFirstChild(xmlDeviceTypeInfo,xmlNodeBrowse));
						if (!mappingHasKey(mScripts, sScriptId))
						{
							dynAppend(dsIgnoreFunctions, substr(sScriptId, strlen(DEVICE_TYPE_CUSTOM_SCRIPT_TAG_START)));
							mScripts[sScriptId] = "//begin_" + sScriptId + "\n" + sScriptContent + "\n" + "//end_" + sScriptId;
						}
						break;
					}
					
					
					case XML_NODE_DEVICE_DPLIST:
					{
						mapping mAttributes = xmlElementAttributes(xmlDeviceTypeInfo, xmlNodeBrowse);
						
						string sDplId;
						
						if (mappingHasKey(mAttributes, XML_NODE_DEVICE_DPLIST_ATTR_NAME))
						{
							sDplId = mAttributes[XML_NODE_DEVICE_DPLIST_ATTR_NAME];
						}
						
						if ("" == sDplId)
						{
							string sError = "Error in file \"" + sFileName + "\"\n" + getCatStr("unDeviceTypeCreator","LOAD-INCORRECTXMLSTRUCTURE2");

							fwException_raise(
									dsExceptions,
									"ERROR",
									sError,
									"LOAD-INCORRECTXMLSTRUCTURE"
								);
							return; 
						}          
						
						
						if (!mappingHasKey(mDpLists, sDplId))
						{
							mDpLists[sDplId] = xmlNodeValue(xmlDeviceTypeInfo, xmlFirstChild(xmlDeviceTypeInfo,xmlNodeBrowse));
						}
						break;
					}
				}
				
				xmlNodeBrowse = xmlNextSibling(xmlDeviceTypeInfo, xmlNodeBrowse);
			}
			
			mapping mTemp;
			mTemp[DEVICE_TYPE_INDEX_SCRIPT] = mScripts;
			mTemp[DEVICE_TYPE_INDEX_DPLIST] = mDpLists;			
			mTemp[DEVICE_TYPE_INDEX_IGNORE_FUNCTIONS] = dsIgnoreFunctions;
			mTemp[DEVICE_TYPE_INDEX_IGNORE_PANELS] = dsIgnorePanels;
			mTemp[DEVICE_TYPE_INDEX_IGNORE_SCRIPTS] = dsIgnoreScripts;
			mTemp[DEVICE_TYPE_INDEX_IGNORE_DPLISTS] = dsIgnoreDpLists;
			mTemp[DEVICE_TYPE_INDEX_IGNORE_IMAGES] = dsIgnoreImages;
			mTemp[DEVICE_TYPE_INDEX_INCLUDE] = dsInclude;

			g_mDeviceTypeInfo[sDeviceType] = mTemp;
		}
	}
	
}

//_unDeviceTypeCreator_stringEndsBy
/**
  @par Description:
  Check if the given string ends like the given string.
  
  @par Usage: 
  Internal
  
	@param sString input, the string in which to look.
	@param sEndsBy input, the string to look for at the end of the previous one
	
	@return true if it matches, false otherwise.
*/
bool _unDeviceTypeCreator_stringEndsBy(string sString, string sEndsBy)
{
	return patternMatch("*" + sEndsBy, sString);
}

//_unDeviceTypeCreator_replaceTags
/**
  @par Description:
  Replaces the tags (e.g. TAG_DeviceType) in the given line by the appropriate value.
  
  @par Usage: 
  Internal
  
	@param sTypeName input, the value to replace tags with.
	@param iTypeType input, indicates wether it is a front-end (CURRENT_TYPE_FRONT_END), a device (CURRENT_TYPE_DEVICE), a widget type (CURRENT_TYPE_WIDGET) or none (CURRENT_TYPE_NONE, results in an error).
	@param sLine    output, the line in which the tags are replaced.
*/
void _unDeviceTypeCreator_replaceTags(string sTypeName, int iType, string &sLine)
{
  strreplace(sLine,TAG_package, g_sPackageName);
  strreplace(sLine,TAG_PACKAGE, strtoupper(g_sPackageName));
  strreplace(sLine,TAG_prefix, g_sPackagePrefix);
  strreplace(sLine,TAG_PREFIX, strtoupper(g_sPackagePrefix));
  
  switch (iType)
  {
    case CURRENT_TYPE_FRONT_END:
      strreplace(sLine,TAG_FrontEndType,sTypeName);
      strreplace(sLine,TAG_FRONTENDTYPE,strtoupper(sTypeName));
      break;
    case CURRENT_TYPE_DEVICE: 
      strreplace(sLine,TAG_DeviceType,sTypeName);
      strreplace(sLine,TAG_DEVICETYPE,strtoupper(sTypeName));
      break;
    case CURRENT_TYPE_WIDGET: 
      strreplace(sLine,TAG_WidgetType,sTypeName);
      strreplace(sLine,TAG_WIDGETTYPE,strtoupper(sTypeName));
    default:
      break;
  }

}

//_unDeviceTypeCreator_generatePackage
/**
  @par Description:
  Generates a package from all previously saved data.
  
  @par Usage: 
  Internal
  
	
	@param dsExceptions output, list of errors that occurred during the process.
*/
int _unDeviceTypeCreator_generatePackage(dyn_string &dsExceptions)
{
	_unDeviceTypeCreator_loadDeviceTypeInfo(dsExceptions);
	  
  // ----------------------------------------- //
  // ------ XML document initialization ------ //
  // ----------------------------------------- //
  
  g_xmlPackageDescriptionFirstChild = xmlAppendChild(g_xmlPackageDescription, -1, XML_ELEMENT_NODE, XML_NODE_COMPONENT);
  
  int xmlNodeIdName = xmlAppendChild(g_xmlPackageDescription, g_xmlPackageDescriptionFirstChild, XML_ELEMENT_NODE,XML_NODE_NAME);
  xmlAppendChild(g_xmlPackageDescription, xmlNodeIdName, XML_TEXT_NODE, g_sPackageName);
  
  int xmlNodeIdVersion = xmlAppendChild(g_xmlPackageDescription,g_xmlPackageDescriptionFirstChild,XML_ELEMENT_NODE,XML_NODE_VERSION);
  xmlAppendChild(g_xmlPackageDescription, xmlNodeIdVersion, XML_TEXT_NODE, g_sNewPackageVersion);
    
  int xmlNodeIdDate = xmlAppendChild(g_xmlPackageDescription,g_xmlPackageDescriptionFirstChild,XML_ELEMENT_NODE,XML_NODE_DATE);
  xmlAppendChild(g_xmlPackageDescription, xmlNodeIdDate, XML_TEXT_NODE, formatTime("%d/%m/%Y",getCurrentTime()));
    
  
  // Export current structure (content of lists and mapping) before any action.
	// Only if the procedure is called from the dedicated panel, otherwise this step is ignored
  // The file is generated at the end.
	unsigned xmlStructure = _unDeviceTypeCreator_exportStructure(g_sPackageName, g_sPackagePrefix);
  
      
  
  // -------------------------------------------------------------------------------- //
  // ------------------------------------ Main loop --------------------------------- //
  // -------------------------------------------------------------------------------- //
  
    
  // Access first child inside "component"  
  unsigned nextSibling = xmlFirstChild(g_xmlTemplateFileList, g_xmlTemplateFileListFirstChild);
  
  // Count children to update the progress bar.
  dyn_int diChildren;
  xmlChildNodes(g_xmlTemplateFileList, xmlParentNode(g_xmlTemplateFileList, nextSibling), diChildren);
  int iChildQty = dynlen(diChildren);
  int iCurrentChildIndex = 0;

  // Iterate through all sibling tags.  
  while (nextSibling != -1)
  {
    iCurrentChildIndex++;
    unsigned xmlNodeIdText = xmlFirstChild(g_xmlTemplateFileList, nextSibling);
    // Contains the path to the current file in the loop
    
    if (xmlNodeType(g_xmlTemplateFileList, nextSibling) != XML_ELEMENT_NODE)
    {
        nextSibling = xmlNextSibling(g_xmlTemplateFileList, nextSibling);
        continue;    
    }
    string sFilePath = xmlNodeValue(g_xmlTemplateFileList, xmlNodeIdText);
    _unDeviceTypeCreator_showProgress("Generating.","Generating file from template.","Parsing " + sFilePath, floor(95*((float)iCurrentChildIndex/(float)iChildQty)));
    // Contains the path to the current file to generate (i.e. the current file after tag replacement
    string sModifiedFilePath = sFilePath;
    // Only replace generic tags for now (package and prefix)
    _unDeviceTypeCreator_replaceTags("",CURRENT_TYPE_NONE,sModifiedFilePath);
    switch (xmlNodeName(g_xmlTemplateFileList,nextSibling))
    {
        
      // ---------------------------------------------------------------- Image ---------------------------------------------------------------- \\
        
      // When the file is an image
      case XML_NODE_IMAGE:
      {
        // Does this image concern a front-end type, a device type, or none ?
        bool bSubTypeConcerned;
        xmlGetElementAttribute(g_xmlTemplateFileList,nextSibling,XML_NODE_ATTRIBUTE_SUBTYPE,bSubTypeConcerned);
        int iTypeConcerned;
        xmlGetElementAttribute(g_xmlTemplateFileList,nextSibling,XML_NODE_ATTRIBUTE_CONCERNS,iTypeConcerned);
  
        // Generic file, does not concern any type
        if (iTypeConcerned == CURRENT_TYPE_NONE)
        {
          // Impossible, an image is always linked to a device or front-end type. 
          // Error in the description file
          // Ignore the file
          break;
        
        }
        mapping mCurrentMapping = _unDeviceTypeCreator_getMappingForType(iTypeConcerned);
        
        for (int i = 1 ; i <= mappinglen(mCurrentMapping) ; i++)
        {
          string sCurrentTypeName = mappingGetKey(mCurrentMapping,i);  
          
					// Image to ignore?
					bool bIgnore = false;
					if (CURRENT_TYPE_DEVICE == iTypeConcerned)
					{
						if (mappingHasKey(g_mDeviceTypeInfo, sCurrentTypeName))
						{
							for (int j = 1 ; j <= dynlen(g_mDeviceTypeInfo[sCurrentTypeName][DEVICE_TYPE_INDEX_IGNORE_IMAGES]) ; j++)
							{
								if (g_mDeviceTypeInfo[sCurrentTypeName][DEVICE_TYPE_INDEX_IGNORE_IMAGES][j] == sFilePath)
								{
									bIgnore = true;
								}
							}
						}
					}					
					if (bIgnore)
					{
						continue;
					}	
					
          // Do nothing if the type already exists and we are not editing a package
          if (_unDeviceTypeCreator_typeExists(sCurrentTypeName,iTypeConcerned,dsExceptions) && !bUpdatingExistingPackage)
          {
            continue;
          }
          if (dynlen(dsExceptions) > 0)
          {
            return -1;
          }
        
          string sTempFilePathType = sModifiedFilePath;
          // Replace the TAG_*Type
          _unDeviceTypeCreator_replaceTags(sCurrentTypeName,iTypeConcerned,sTempFilePathType);
              
          // Need to duplicate the image ?
          if (bSubTypeConcerned)
          {
            for (int i = 1 ; i <= dynlen(mCurrentMapping[sCurrentTypeName]) ; i++)
            {
              string sTempFilePathSubType = sTempFilePathType;
              _unDeviceTypeCreator_replaceTags(mCurrentMapping[sCurrentTypeName][i],iTypeConcerned-1,sTempFilePathSubType);
        
              if (isfile(g_sTemplatePath + DEVICE_TYPE_INFO_FOLDER + sTempFilePathSubType))
              {
                bool bFileCopied = copyFile(g_sTemplatePath + DEVICE_TYPE_INFO_FOLDER + sTempFilePathSubType, g_sNewPackagePath + "/" + g_sPackageName + "/" + sTempFilePathSubType);
                if (!bFileCopied)
                {
                  string sError = getCatStr("unDeviceTypeCreator","GENERATE-COPYERROR");
                  strreplace(sError,"SOURCE",g_sTemplatePath + "/" +DEVICE_TYPE_INFO_FOLDER + sTempFilePathSubType);
                  strreplace(sError,"DESTINATION",g_sNewPackagePath + "/" + g_sPackageName + "/" + sTempFilePathSubType);
      
                  fwException_raise(
                      dsExceptions,
                      "ERROR",
                      sError,
                      "GENERATE-COPYERROR"
                  );
                  return -1;
                }
              } 
              else
              // If the file already exists copy this one instead of overwriting it
              if (_unDeviceTypeCreator_fileExists(sTempFilePathSubType))
              {
                if (!copyFile(mLoadedPackageLocation + "/" + sTempFilePathSubType, g_sNewPackagePath + "/" + g_sPackageName + "/" + sTempFilePathSubType))
                {          
                  string sError = getCatStr("unDeviceTypeCreator","GENERATE-COPYERROR");
                  strreplace(sError,"SOURCE",mLoadedPackageLocation + "/" + sTempFilePathSubType);
                  strreplace(sError,"DESTINATION",g_sNewPackagePath + "/" + g_sPackageName + "/" + sTempFilePathSubType);
      
                  fwException_raise(
                      dsExceptions,
                      "ERROR",
                      sError,
                      "GENERATE-COPYERROR"
                  );
                  return -1;
                }
              } 
              else if (!copyFile(g_sTemplatePath + "/" + sFilePath, g_sNewPackagePath + "/" + g_sPackageName + "/" + sTempFilePathSubType))
              {
                string sError = getCatStr("unDeviceTypeCreator","GENERATE-COPYERROR");
                strreplace(sError,"SOURCE",g_sTemplatePath + "/" + sFilePath);
                strreplace(sError,"DESTINATION",g_sNewPackagePath + "/" + g_sPackageName + "/" + sTempFilePathSubType);
      
                fwException_raise(
                    dsExceptions,
                    "ERROR",
                    sError,
                    "GENERATE-COPYERROR"
                );
                return -1;
              }
              
              // Add to the XML component file
              int xmlNodeIdFile = xmlAppendChild(g_xmlPackageDescription,g_xmlPackageDescriptionFirstChild,XML_ELEMENT_NODE,XML_NODE_FILE);
              xmlAppendChild(g_xmlPackageDescription,xmlNodeIdFile,XML_TEXT_NODE,sTempFilePathSubType);   
            } 
            
          }
          else
          {
            if (isfile(g_sTemplatePath + DEVICE_TYPE_INFO_FOLDER + sTempFilePathType))
            {
              bool bFileCopied = copyFile(g_sTemplatePath + DEVICE_TYPE_INFO_FOLDER + sTempFilePathType, g_sNewPackagePath + "/" + g_sPackageName + "/" + sTempFilePathType);
              if (!bFileCopied)
              {
                string sError = getCatStr("unDeviceTypeCreator","GENERATE-COPYERROR");
                strreplace(sError,"SOURCE",g_sTemplatePath + "/" +DEVICE_TYPE_INFO_FOLDER + sTempFilePathType);
                strreplace(sError,"DESTINATION",g_sNewPackagePath + "/" + g_sPackageName + "/" + sTempFilePathType);
      
                fwException_raise(
                    dsExceptions,
                    "ERROR",
                    sError,
                    "GENERATE-COPYERROR"
                );
                return -1;
              }
            } 
            else
            // If the file already exists copy this one instead of overwriting it
            if (_unDeviceTypeCreator_fileExists(sTempFilePathType))
            {
              if (!copyFile(mLoadedPackageLocation + "/" + sTempFilePathType, g_sNewPackagePath + "/" + g_sPackageName + "/" + sTempFilePathType))
              {          
                string sError = getCatStr("unDeviceTypeCreator","GENERATE-COPYERROR");
                strreplace(sError,"SOURCE",mLoadedPackageLocation + "/" + sTempFilePathType);
                strreplace(sError,"DESTINATION",g_sNewPackagePath + "/" + g_sPackageName + "/" + sTempFilePathType);
      
                fwException_raise(
                    dsExceptions,
                    "ERROR",
                    sError,
                    "GENERATE-COPYERROR"
                );
                return -1;
              }
            }
            else if (!copyFile(g_sTemplatePath + "/" + sFilePath, g_sNewPackagePath + "/" + g_sPackageName + "/" + sTempFilePathType))
            {          
              string sError = getCatStr("unDeviceTypeCreator","GENERATE-COPYERROR");
              strreplace(sError,"SOURCE",g_sTemplatePath + "/" + sFilePath);
              strreplace(sError,"DESTINATION",g_sNewPackagePath + "/" + g_sPackageName + "/" + sTempFilePathType);
      
              fwException_raise(
                  dsExceptions,
                  "ERROR",
                  sError,
                  "GENERATE-COPYERROR"
              );
              return -1;
            }
            
            // Add to the XML component file
            int xmlNodeIdFile = xmlAppendChild(g_xmlPackageDescription,g_xmlPackageDescriptionFirstChild,XML_ELEMENT_NODE,XML_NODE_FILE);
            xmlAppendChild(g_xmlPackageDescription,xmlNodeIdFile,XML_TEXT_NODE,sTempFilePathType);   
          }
        }
        break;        
      }
      
        
      // ---------------------------------------------------------------- Panel ---------------------------------------------------------------- \\
        
      // When it is a panel.
      case XML_NODE_PANEL:
      {
        // Does this panel concern a front-end type, a device type, or none ?
        bool bSubTypeConcerned;
        xmlGetElementAttribute(g_xmlTemplateFileList,nextSibling,XML_NODE_ATTRIBUTE_SUBTYPE,bSubTypeConcerned);
        int iTypeConcerned;
        xmlGetElementAttribute(g_xmlTemplateFileList,nextSibling,XML_NODE_ATTRIBUTE_CONCERNS,iTypeConcerned);
        int iParticularCase = 0;
        xmlGetElementAttribute(g_xmlTemplateFileList,nextSibling,XML_NODE_ATTRIBUTE_CASE,iParticularCase);
        
			
        mapping mCurrentMapping = _unDeviceTypeCreator_getMappingForType(iTypeConcerned);
        
        // Some panels may be unique and cannot be treated like the others.
        // This cases are identified by known integers.
        // When we encounter one of these cases, we need to proceed particularly instead of following the habitual program.

        // Particular case #1 : TAG_prefixTestWidgets.pnl
        // Note : if the panel already exists, it is overwritten and former content is lost.
        if (iParticularCase == 1)
        {            
          if (isfile(g_sTemplatePath + DEVICE_TYPE_INFO_FOLDER + sModifiedFilePath))
          {
            bool bFileCopied = copyFile(g_sTemplatePath + DEVICE_TYPE_INFO_FOLDER + sModifiedFilePath, g_sNewPackagePath + "/" + g_sPackageName + "/" + sModifiedFilePath);
            if (!bFileCopied)
            {
              string sError = getCatStr("unDeviceTypeCreator","GENERATE-COPYERROR");
              strreplace(sError,"SOURCE",g_sTemplatePath + "/" +DEVICE_TYPE_INFO_FOLDER + sModifiedFilePath);
              strreplace(sError,"DESTINATION",g_sNewPackagePath + "/" + g_sPackageName + "/" + sModifiedFilePath);
      
              fwException_raise(
                  dsExceptions,
                  "ERROR",
                  sError,
                  "GENERATE-COPYERROR"
              );
              return -1;
            }            
          }  
          else
          {
            string sDeviceList;
            for (int i = 1 ; i <= mappinglen(mCurrentMapping) ; i++)
            {
              sDeviceList += "\\\"" + mappingGetKey(mCurrentMapping,i) + "\\\"";
              if (i <  mappinglen(mCurrentMapping))
              {
                sDeviceList += ",";
              }
            } 
            
            string sPanelContent;
            fileToString(g_sTemplatePath + sFilePath,sPanelContent);
            strreplace(sPanelContent,"\\\"TAG_DeviceType\\\"",sDeviceList);
          
            file f = fopen(g_sTemplatePath + "/" + baseName(sModifiedFilePath),"w");
            fputs(sPanelContent,f);
            fclose(f);
          
            bool bFileCopied = copyFile(g_sTemplatePath + "/" + baseName(sModifiedFilePath), g_sNewPackagePath + "/" + g_sPackageName + "/" + sModifiedFilePath);
            remove(g_sTemplatePath + "/" + baseName(sModifiedFilePath));
            if (!bFileCopied)
            {
              string sError = getCatStr("unDeviceTypeCreator","GENERATE-COPYERROR");
              strreplace(sError,"SOURCE", g_sTemplatePath + "/" + baseName(sModifiedFilePath));
              strreplace(sError,"DESTINATION", g_sNewPackagePath + "/" + g_sPackageName + "/" + sModifiedFilePath);
      
              fwException_raise(
                  dsExceptions,
                  "ERROR",
                  sError,
                  "GENERATE-COPYERROR"
              );
              return -1;
            }
          }
          // Add to the XML component file
          int xmlNodeIdFile = xmlAppendChild(g_xmlPackageDescription,g_xmlPackageDescriptionFirstChild,XML_ELEMENT_NODE,XML_NODE_FILE);
          xmlAppendChild(g_xmlPackageDescription,xmlNodeIdFile,XML_TEXT_NODE,sModifiedFilePath); 
            
          
          break;
        }
        
        string sTagToReplace;
        switch (iTypeConcerned)
        {
          case CURRENT_TYPE_FRONT_END:
            sTagToReplace = TAG_FrontEndType;
            break;
          case CURRENT_TYPE_DEVICE:
            sTagToReplace = TAG_DeviceType;
            break;
          case CURRENT_TYPE_WIDGET:
            sTagToReplace = TAG_WidgetType;
            break;
          default:
            sTagToReplace = "";
            break;        
        }
        
        if (sTagToReplace == "")
        {
          break;
        }
        
        for (int i = 1 ; i <= mappinglen(mCurrentMapping) ; i++)
        {
          string sCurrentTypeName = mappingGetKey(mCurrentMapping,i);    
          
					
				
					// Panel to ignore?
					bool bIgnore = false;
					if (CURRENT_TYPE_DEVICE == iTypeConcerned)
					{
						if (mappingHasKey(g_mDeviceTypeInfo, sCurrentTypeName))
						{
							for (int j = 1 ; j <= dynlen(g_mDeviceTypeInfo[sCurrentTypeName][DEVICE_TYPE_INDEX_IGNORE_PANELS]) ; j++)
							{
								if (g_mDeviceTypeInfo[sCurrentTypeName][DEVICE_TYPE_INDEX_IGNORE_PANELS][j] == sFilePath)
								{
									bIgnore = true;
								}
							}
						}
					}					
					if (bIgnore)
					{
						continue;
					}					
			
          // Do nothing if the type already exists and we are not editing an existing package
          if (_unDeviceTypeCreator_typeExists(sCurrentTypeName,iTypeConcerned,dsExceptions) && !bUpdatingExistingPackage)
          {
            continue;
          }
          if (dynlen(dsExceptions) > 0)
          {
            return -1;
          }
           
          
           
          string sTempFilePathType = sModifiedFilePath;
          // Replace the TAG_*Type
          _unDeviceTypeCreator_replaceTags(sCurrentTypeName,iTypeConcerned,sTempFilePathType);
          
          
          // Warning : some panels contain an indicator of size of string (e.g. LANG:1 16 TAG_DeviceType), in this case we must also change the size.
          // Problem : how to do this when PVSS does not have proper regular expression functions...
          // 1) Extract panel content          
          string sPanelContent;
          fileToString(g_sTemplatePath + sFilePath,sPanelContent);
          // 2) Split the panel lines
          dyn_string dsSplittedPanelContent = strsplit(sPanelContent,"\n");
          // 3) Use patternMatch(), the closest function to regular expressions          
          dyn_bool dbPatternMatch = patternMatch("LANG:1 ?? " + sTagToReplace + "*",dsSplittedPanelContent);
          int indexPatternMatches = dynContains(dbPatternMatch,true);
          if (indexPatternMatches > 0)
          {
            // 4) One line matches the pattern, we need to replace its content
            string sMatchingLine = dsSplittedPanelContent[indexPatternMatches];  
            strreplace(sMatchingLine,sTagToReplace,sCurrentTypeName);
            strreplace(
                sPanelContent,
                dsSplittedPanelContent[indexPatternMatches],
                "LANG:1 " + strlen(substr(sMatchingLine,strpos(sMatchingLine,sCurrentTypeName))) + " " + substr(sMatchingLine,strpos(sMatchingLine,sCurrentTypeName))
              );          
          }
          // Do not forget to replace everything else inside the file and file name too.
          _unDeviceTypeCreator_replaceTags(sCurrentTypeName,iTypeConcerned,sPanelContent);
          
          // Duplicate ?
          if (bSubTypeConcerned)
          {
            for (int i = 1 ; i <= dynlen(mCurrentMapping[sCurrentTypeName]) ; i++)
            {
              string sTempPanelContent = sPanelContent;
              _unDeviceTypeCreator_replaceTags(mCurrentMapping[sCurrentTypeName][i], iTypeConcerned - 1, sTempPanelContent);
              string sTempFileName = sTempFilePathType;
              _unDeviceTypeCreator_replaceTags(mCurrentMapping[sCurrentTypeName][i], iTypeConcerned - 1, sTempFileName);
            
              if (isfile(g_sTemplatePath + DEVICE_TYPE_INFO_FOLDER + sTempFileName))
              {
                bool bFileCopied = copyFile(g_sTemplatePath + DEVICE_TYPE_INFO_FOLDER + sTempFileName, g_sNewPackagePath + "/" + g_sPackageName + "/" + sTempFileName);
                if (!bFileCopied)
                {
                  string sError = getCatStr("unDeviceTypeCreator","GENERATE-COPYERROR");
                  strreplace(sError,"SOURCE",g_sTemplatePath + "/" + DEVICE_TYPE_INFO_FOLDER + sTempFileName);
                  strreplace(sError,"DESTINATION",g_sNewPackagePath + "/" + g_sPackageName + "/" + sTempFileName);
      
                  fwException_raise(
                      dsExceptions,
                      "ERROR",
                      sError,
                      "GENERATE-COPYERROR"
                  );
                  return -1;
                }
              }
              else
              // If the file already exists copy this one instead of overwriting it
              if (_unDeviceTypeCreator_fileExists(sTempFileName))
              {
                if (!copyFile(mLoadedPackageLocation + "/" + sTempFileName, g_sNewPackagePath + "/" + g_sPackageName + "/" + sTempFileName))
                {          
                  string sError = getCatStr("unDeviceTypeCreator","GENERATE-COPYERROR");
                  strreplace(sError,"SOURCE",mLoadedPackageLocation + "/" + sTempFileName);
                  strreplace(sError,"DESTINATION",g_sNewPackagePath + "/" + g_sPackageName + "/" + sTempFileName);
      
                  fwException_raise(
                      dsExceptions,
                      "ERROR",
                      sError,
                      "GENERATE-COPYERROR"
                  );
                  return -1;
                }
              } 
              else
              {              
                // fopen() cannot create subdirectories. We have to create the file in the template root folder then use copyfile() which creates subdirs
                file f = fopen(g_sTemplatePath + "/" + baseName(sTempFileName),"w");
                fputs(sTempPanelContent,f);
                fclose(f);
                bool bFileCopied = copyFile(g_sTemplatePath + "/" + baseName(sTempFileName), g_sNewPackagePath + "/" + g_sPackageName + "/" + sTempFileName);
                remove(g_sTemplatePath + "/" + baseName(sTempFileName));
                if (!bFileCopied)
                {
                  string sError = getCatStr("unDeviceTypeCreator","GENERATE-COPYERROR");
                  strreplace(sError,"SOURCE",g_sTemplatePath + "/" + baseName(sTempFileName));
                  strreplace(sError,"DESTINATION",g_sNewPackagePath + "/" + g_sPackageName + "/" + sTempFileName);
      
                  fwException_raise(
                      dsExceptions,
                      "ERROR",
                      sError,
                      "GENERATE-COPYERROR"
                  );
                  return -1;
                }
              }
            
              // Add to the XML component file
              int xmlNodeIdFile = xmlAppendChild(g_xmlPackageDescription,g_xmlPackageDescriptionFirstChild,XML_ELEMENT_NODE,XML_NODE_FILE);
              xmlAppendChild(g_xmlPackageDescription,xmlNodeIdFile,XML_TEXT_NODE,sTempFileName);    
            }
          }
          else
          {
            if (isfile(g_sTemplatePath + DEVICE_TYPE_INFO_FOLDER + sTempFilePathType))
            {
              bool bFileCopied = copyFile(g_sTemplatePath + DEVICE_TYPE_INFO_FOLDER + sTempFilePathType, g_sNewPackagePath + "/" + g_sPackageName + "/" + sTempFilePathType);
              if (!bFileCopied)
              {
                string sError = getCatStr("unDeviceTypeCreator","GENERATE-COPYERROR");
                strreplace(sError,"SOURCE",g_sTemplatePath + "/" + DEVICE_TYPE_INFO_FOLDER + sTempFilePathType);
                strreplace(sError,"DESTINATION",g_sNewPackagePath + "/" + g_sPackageName + "/" + sTempFilePathType);
      
                fwException_raise(
                    dsExceptions,
                    "ERROR",
                    sError,
                    "GENERATE-COPYERROR"
                );
                return -1;
              }
            }
            // If the file already exists copy this one instead of overwriting it
            else if (_unDeviceTypeCreator_fileExists(sTempFilePathType))
            {
              if (!copyFile(mLoadedPackageLocation + "/" + sTempFilePathType, g_sNewPackagePath + "/" + g_sPackageName + "/" + sTempFilePathType))
              {          
                string sError = getCatStr("unDeviceTypeCreator","GENERATE-COPYERROR");
                strreplace(sError,"SOURCE",mLoadedPackageLocation + "/" + sTempFilePathType);
                strreplace(sError,"DESTINATION",g_sNewPackagePath + "/" + g_sPackageName + "/" + sTempFilePathType);
      
                fwException_raise(
                    dsExceptions,
                    "ERROR",
                    sError,
                    "GENERATE-COPYERROR"
                );
                return -1;
              }
            }  
            else
            {
              // fopen() cannot create subdirectories. We have to create the file in the template root folder then use copyfile() which creates subdirs
              file f = fopen(g_sTemplatePath + "/" + baseName(sTempFilePathType),"w");
              fputs(sPanelContent,f);
              fclose(f);
              bool bFileCopied = copyFile(g_sTemplatePath + "/" + baseName(sTempFilePathType), g_sNewPackagePath + "/" + g_sPackageName + "/" + sTempFilePathType);
              remove(g_sTemplatePath + "/" + baseName(sTempFilePathType));
              if (!bFileCopied)
              {
                string sError = getCatStr("unDeviceTypeCreator","GENERATE-COPYERROR");
                strreplace(sError,"SOURCE", g_sTemplatePath + "/" + baseName(sTempFilePathType));
                strreplace(sError,"DESTINATION", g_sNewPackagePath + "/" + g_sPackageName + "/" + sTempFilePathType);
      
                fwException_raise(
                    dsExceptions,
                    "ERROR",
                    sError,
                    "GENERATE-COPYERROR"
                );
                return -1;
              }            
            }
          
            // Add to the XML component file
            int xmlNodeIdFile = xmlAppendChild(g_xmlPackageDescription,g_xmlPackageDescriptionFirstChild,XML_ELEMENT_NODE,XML_NODE_FILE);
            xmlAppendChild(g_xmlPackageDescription,xmlNodeIdFile,XML_TEXT_NODE,sTempFilePathType); 
          }
        }
        break;
      }
        
      // ---------------------------------------------------------------- Script ---------------------------------------------------------------- \\
        
      // Tricky part : when the file is a CTRL script
      case XML_NODE_SCRIPT:    
      {
        // Parse XML File
        string errMsg;
        int errLine;
        int errColumn;
        unsigned scriptXmlDescription = xmlDocumentFromFile(g_sTemplatePath + "/" + sFilePath,errMsg,errLine,errColumn);
        // The template is supposed to be correct (version check), so no need to check anything. If it not correct it is the user's fault.
        
        // Access the first child of the document
        unsigned ctrlScriptNode = xmlFirstChild(scriptXmlDescription);
        // Some files may contain the tag <?xml/>, in that case we must reach the next element
        if (xmlNodeName(scriptXmlDescription, ctrlScriptNode) != XML_TEMPLATE_CTRLSCRIPT)
        {
          ctrlScriptNode = xmlNextSibling(scriptXmlDescription, ctrlScriptNode);
        }  
        
        // Does this script concern a front-end type, a device type, or none ?
        int iTypeConcerned;
        xmlGetElementAttribute(scriptXmlDescription,ctrlScriptNode,XML_NODE_ATTRIBUTE_CONCERNS,iTypeConcerned);
        
        bool bConfigUiOnly;
        xmlGetElementAttribute(g_xmlTemplateFileList,nextSibling,XML_NODE_ATTRIBUTE_UIONLY,bConfigUiOnly);
        
        
        // Some generic scripts don't need to be duplicated.
        if (iTypeConcerned == CURRENT_TYPE_NONE)
        {
          if (!_unDeviceTypeCreator_fileExists(sModifiedFilePath))
          {
            string sNewFileContent = _unDeviceTypeCreator_createNewTypeScript("", iTypeConcerned, scriptXmlDescription, ctrlScriptNode, dsExceptions, sModifiedFilePath);
            strreplace(sNewFileContent,"\r", "");
            _unDeviceTypeCreator_replaceTags("",CURRENT_TYPE_NONE,sNewFileContent);
          
          
            file f = fopen(g_sTemplatePath + "/" + baseName(sModifiedFilePath),"w");
            fputs(sNewFileContent,f);
            fclose(f);
          
            string sConfigLibPath = substr(sModifiedFilePath, strpos(sModifiedFilePath, "/libs/") + strlen("/libs/"));
            if (bConfigUiOnly)
            {
              dynAppend(g_dsLibListUiOnly, sConfigLibPath);
            }
            else
            {
              dynAppend(g_dsLibList, sConfigLibPath);
            }
          
            bool bFileCopied = copyFile(g_sTemplatePath + "/" + baseName(sModifiedFilePath), g_sNewPackagePath + "/" + g_sPackageName + "/" + sModifiedFilePath);
            remove(g_sTemplatePath + "/" + baseName(sModifiedFilePath));
            if (!bFileCopied)
            {
              string sError = getCatStr("unDeviceTypeCreator","GENERATE-COPYERROR");
              strreplace(sError,"SOURCE",g_sTemplatePath + "/" + baseName(sModifiedFilePath));
              strreplace(sError,"DESTINATION",g_sNewPackagePath + "/" + g_sPackageName + "/" + sModifiedFilePath);
      
              fwException_raise(
                  dsExceptions,
                  "ERROR",
                  sError,
                  "GENERATE-COPYERROR"
              );
              return -1;
            }
          }
          else
          {
            bool bFileCopied = copyFile(mLoadedPackageLocation + "/" + sModifiedFilePath, g_sNewPackagePath + "/" + g_sPackageName + "/" + sModifiedFilePath);

            string sConfigLibPath = substr(sModifiedFilePath, strpos(sModifiedFilePath, "/libs/") + strlen("/libs/"));
            if (bConfigUiOnly)
            {
              dynAppend(g_dsLibListUiOnly, sConfigLibPath);
            }
            else
            {
              dynAppend(g_dsLibList, sConfigLibPath);
            }
            
            if (!bFileCopied)
            {
              string sError = getCatStr("unDeviceTypeCreator","GENERATE-COPYERROR");
              strreplace(sError,"SOURCE",mLoadedPackageLocation + "/" + sModifiedFilePath);
              strreplace(sError,"DESTINATION",g_sNewPackagePath + "/" + g_sPackageName + "/" + sModifiedFilePath);
      
              fwException_raise(
                  dsExceptions,
                  "ERROR",
                  sError,
                  "GENERATE-COPYERROR"
              );
              return -1;
            }
          }
          
          // Add to the XML component file
          int xmlNodeIdFile = xmlAppendChild(g_xmlPackageDescription,g_xmlPackageDescriptionFirstChild,XML_ELEMENT_NODE,XML_NODE_FILE);
          xmlAppendChild(g_xmlPackageDescription,xmlNodeIdFile,XML_TEXT_NODE,sModifiedFilePath);
                    
          break;
        }
        
        
        mapping mCurrentMapping = _unDeviceTypeCreator_getMappingForType(iTypeConcerned);
        
        for (int i = 1 ; i <= mappinglen(mCurrentMapping) ; i++)
        {
          string sCurrentTypeName = mappingGetKey(mCurrentMapping,i);     
          
					// Script to ignore?
					bool bIgnore = false;
					if (CURRENT_TYPE_DEVICE == iTypeConcerned)
					{
						if (mappingHasKey(g_mDeviceTypeInfo, sCurrentTypeName))
						{
							for (int j = 1 ; j <= dynlen(g_mDeviceTypeInfo[sCurrentTypeName][DEVICE_TYPE_INDEX_IGNORE_SCRIPTS]) ; j++)
							{
								if (g_mDeviceTypeInfo[sCurrentTypeName][DEVICE_TYPE_INDEX_IGNORE_SCRIPTS][j] == sFilePath)
								{
									bIgnore = true;
								}
							}
						}
					}					
					if (bIgnore)
					{
						continue;
					}
					
					
          // Do nothing if the type already exists and we are not editing a package
          if (_unDeviceTypeCreator_typeExists(sCurrentTypeName,iTypeConcerned,dsExceptions) && !bUpdatingExistingPackage)
          {
            continue;
          }
          if (dynlen(dsExceptions) > 0)
          {
            return -1;
          }
           
          
          string sTempFilePathType = sModifiedFilePath;
          _unDeviceTypeCreator_replaceTags(sCurrentTypeName,iTypeConcerned,sTempFilePathType);
          string sNewFileContent = _unDeviceTypeCreator_createNewTypeScript(sCurrentTypeName, iTypeConcerned, scriptXmlDescription, ctrlScriptNode, dsExceptions, sTempFilePathType);
          strreplace(sNewFileContent,"\r", "");
          
          file f = fopen(g_sTemplatePath + "/" + baseName(sTempFilePathType),"w");
          fputs(sNewFileContent,f);
          fclose(f);
          
          string sConfigLibPath = substr(sTempFilePathType, strpos(sTempFilePathType, "/libs/") + strlen("/libs/"));
          if (bConfigUiOnly)
          {
            dynAppend(g_dsLibListUiOnly, sConfigLibPath);
          }
          else
          {
            dynAppend(g_dsLibList, sConfigLibPath);
          }
          
          bool bFileCopied = copyFile(g_sTemplatePath + "/" + baseName(sTempFilePathType), g_sNewPackagePath + "/" + g_sPackageName + "/" + sTempFilePathType);
          remove(g_sTemplatePath + "/" + baseName(sTempFilePathType));
          if (!bFileCopied)
          {
            string sError = getCatStr("unDeviceTypeCreator","GENERATE-COPYERROR");
            strreplace(sError,"SOURCE",g_sTemplatePath + "/" + baseName(sTempFilePathType));
            strreplace(sError,"DESTINATION",g_sNewPackagePath + "/" + g_sPackageName + "/" + sTempFilePathType);
      
            fwException_raise(
                dsExceptions,
                "ERROR",
                sError,
                "GENERATE-COPYERROR"
            );
            return -1;
          }
          
          // Add to the XML component file
          int xmlNodeIdFile = xmlAppendChild(g_xmlPackageDescription,g_xmlPackageDescriptionFirstChild,XML_ELEMENT_NODE,XML_NODE_FILE);
          xmlAppendChild(g_xmlPackageDescription,xmlNodeIdFile,XML_TEXT_NODE,sTempFilePathType);
          
        }
           
         
        
        break; 
          
          
      }
        
      // ---------------------------------------------------------------- Installation scripts ---------------------------------------------------------------- \\
      
      // Nothing particular to do, just copy/replace.
      case XML_NODE_PREINIT:
      case XML_NODE_INIT:
      case XML_NODE_POSTINSTALL:
      {
        
        int xmlNodeIdFile = xmlAppendChild(g_xmlPackageDescription,g_xmlPackageDescriptionFirstChild,XML_ELEMENT_NODE,xmlNodeName(g_xmlTemplateFileList,nextSibling));
        xmlAppendChild(g_xmlPackageDescription,xmlNodeIdFile,XML_TEXT_NODE,sModifiedFilePath);
        
        if (isfile(g_sTemplatePath + DEVICE_TYPE_INFO_FOLDER + sModifiedFilePath))
        {
          bool bFileCopied = copyFile(g_sTemplatePath + DEVICE_TYPE_INFO_FOLDER + sModifiedFilePath, g_sNewPackagePath + "/" + g_sPackageName + "/" + sModifiedFilePath);
          if (!bFileCopied)
          {
            string sError = getCatStr("unDeviceTypeCreator","GENERATE-COPYERROR");
            strreplace(sError,"SOURCE",g_sTemplatePath + "/" +DEVICE_TYPE_INFO_FOLDER + sModifiedFilePath);
            strreplace(sError,"DESTINATION",g_sNewPackagePath + "/" + g_sPackageName + "/" + sModifiedFilePath);
      
            fwException_raise(
                dsExceptions,
                "ERROR",
                sError,
                "GENERATE-COPYERROR"
            );
            return -1;
          }
          
          break;
        }  
        
        string sFileContent;
        fileToString(g_sTemplatePath + sFilePath,sFileContent);
        _unDeviceTypeCreator_replaceTags("", CURRENT_TYPE_NONE, sFileContent);
        
        file f = fopen(g_sTemplatePath + "/" + baseName(sModifiedFilePath),"w");
        fputs(sFileContent,f);
        fclose(f);
        
        bool bFileCopied = copyFile(g_sTemplatePath + "/" + baseName(sModifiedFilePath), g_sNewPackagePath + "/" + g_sPackageName + "/" + sModifiedFilePath);
        remove(g_sTemplatePath + "/" + baseName(sModifiedFilePath));
        if (!bFileCopied)
        {
          string sError = getCatStr("unDeviceTypeCreator","GENERATE-COPYERROR");
          strreplace(sError,"SOURCE",g_sTemplatePath + "/" + baseName(sModifiedFilePath));
          strreplace(sError,"DESTINATION",g_sNewPackagePath + "/" + g_sPackageName + "/" + sModifiedFilePath);
      
          fwException_raise(
              dsExceptions,
              "ERROR",
              sError,
              "GENERATE-COPYERROR"
          );
          return -1;
        }
        
        break;
      }

      // ---------------------------------------------------------------- DP list ---------------------------------------------------------------- \\
        
      case XML_NODE_DPLIST:
      {
        // Does this file concern a front-end type, a device type, or none ?
        int iTypeConcerned;
        xmlGetElementAttribute(g_xmlTemplateFileList, nextSibling, XML_NODE_ATTRIBUTE_CONCERNS, iTypeConcerned);
        string sFileContent;
        fileToString(g_sTemplatePath + sFilePath,sFileContent);
                
        mapping mCurrentMapping = _unDeviceTypeCreator_getMappingForType(iTypeConcerned);
        for (int i = 1 ; i <= mappinglen(mCurrentMapping) ; i++)
        {
          string sCurrentTypeName = mappingGetKey(mCurrentMapping,i);  
          
					
					// Image to ignore?
					bool bIgnore = false;
					if (CURRENT_TYPE_DEVICE == iTypeConcerned)
					{
						if (mappingHasKey(g_mDeviceTypeInfo, sCurrentTypeName))
						{
							for (int j = 1 ; j <= dynlen(g_mDeviceTypeInfo[sCurrentTypeName][DEVICE_TYPE_INDEX_IGNORE_DPLISTS]) ; j++)
							{
								if (g_mDeviceTypeInfo[sCurrentTypeName][DEVICE_TYPE_INDEX_IGNORE_DPLISTS][j] == sFilePath)
								{
									bIgnore = true;
								}
							}
						}
					}					
					if (bIgnore)
					{
						continue;
					}	
					
					
          // Do nothing if the type already exists since all DP/DPT have already been created.
          if (_unDeviceTypeCreator_typeExists(sCurrentTypeName, iTypeConcerned, dsExceptions))
          {
            continue;
          }          
          if (dynlen(dsExceptions) > 0)
          {
            return -1;
          }
           
          string sTempModifiedFilePath = sModifiedFilePath;
          _unDeviceTypeCreator_replaceTags(sCurrentTypeName, iTypeConcerned, sTempModifiedFilePath);

          
          // If this file already exists in the device type info folder, copy this one instead of creating a new one
          if (isfile(g_sTemplatePath + DEVICE_TYPE_INFO_FOLDER + sTempModifiedFilePath))
          {
            bool bFileCopied = copyFile(g_sTemplatePath + DEVICE_TYPE_INFO_FOLDER + sTempModifiedFilePath, g_sNewPackagePath + "/" + g_sPackageName + "/" + sTempModifiedFilePath);
            if (!bFileCopied)
            {
              string sError = getCatStr("unDeviceTypeCreator","GENERATE-COPYERROR");
              strreplace(sError,"SOURCE", g_sTemplatePath + DEVICE_TYPE_INFO_FOLDER + sTempModifiedFilePath);
              strreplace(sError,"DESTINATION", g_sNewPackagePath + "/" + g_sPackageName + "/" + sTempModifiedFilePath);
      
              fwException_raise(
                  dsExceptions,
                  "ERROR",
                  sError,
                  "GENERATE-COPYERROR"
              );
              return -1;
            }
          }          
          else
          {
            string sTempFileContent = sFileContent;

            //  /!\ Warning. There is a particular case in one of the DPL files : the line containing the list of the widgets of a device.
            // It needs to be correctly corrected.
            if (strpos(sTempFileContent, TAG_WidgetType) >= 0)
            {
              // Particular case found. The line is complex to replace, the exact line to replace must be extracted manually
              // First, look for "unWidgets.List"
              string sWidgetListTemplate = substr(sTempFileContent, strpos(sTempFileContent, "unWidgets.List"));
              // Then look for the first '"'
              sWidgetListTemplate = substr(sWidgetListTemplate, strpos(sWidgetListTemplate, "\"") + 1);
              // Then look for the next '"'
              sWidgetListTemplate = substr(sWidgetListTemplate, 0, strpos(sWidgetListTemplate, "\""));
              // Add quote at start and end            
              sWidgetListTemplate = "\"" + sWidgetListTemplate + "\"";
            
              // The template is ready to be replaced            
            
              string sWidgetList;
              for (int j = 1 ; j <= dynlen(mCurrentMapping[sCurrentTypeName]) ; j++)
              {
                // sWidgetList += "\"objects/TAG_PACKAGE/TAG_prefixWidget_TAG_DeviceType" + mCurrentMapping[sCurrentTypeName][j] + ".pnl\"";
                string sWidgetTemplateTemp = sWidgetListTemplate;
                strreplace(sWidgetTemplateTemp, TAG_WidgetType, mCurrentMapping[sCurrentTypeName][j]);
              
                sWidgetList += sWidgetTemplateTemp;
              
                if (j < dynlen(mCurrentMapping[sCurrentTypeName]))
                {
                  // Append the coma only if not at the last element
                  sWidgetList += ",";
                }
              }
            
              // strreplace(sTempFileContent,"\"objects/TAG_PACKAGE/TAG_prefixWidget_TAG_DeviceTypeTAG_WidgetType.pnl\"",sWidgetList);
              strreplace(sTempFileContent,sWidgetListTemplate, sWidgetList);
            }
          
            _unDeviceTypeCreator_replaceTags(sCurrentTypeName, iTypeConcerned, sTempFileContent);
            
            // Replace all special device type information (if the script is for a device)
            if (CURRENT_TYPE_DEVICE == iTypeConcerned)
            {
              if (mappingHasKey(g_mDeviceTypeInfo, sCurrentTypeName))
              {
                for (int i = 1 ; i <= mappinglen(g_mDeviceTypeInfo[sCurrentTypeName][DEVICE_TYPE_INDEX_DPLIST]) ; i++)
                {
                  string sTag = mappingGetKey(g_mDeviceTypeInfo[sCurrentTypeName][DEVICE_TYPE_INDEX_DPLIST], i);
                  string sContent = mappingGetValue(g_mDeviceTypeInfo[sCurrentTypeName][DEVICE_TYPE_INDEX_DPLIST], i);
        
                  strreplace(sContent, "\r", "");
                  sContent = strrtrim(sContent);
                  sContent = strltrim(sContent, "\n");
                  strreplace(sTempFileContent, sTag, sContent);
                }
              }
            }
          
            file f = fopen(g_sTemplatePath + "/" + baseName(sTempModifiedFilePath),"w");
            fputs(sTempFileContent,f);
            fclose(f);
          
            bool bFileCopied = copyFile(g_sTemplatePath + "/" + baseName(sTempModifiedFilePath), g_sNewPackagePath + "/" + g_sPackageName + "/" + sTempModifiedFilePath);
            remove(g_sTemplatePath + "/" + baseName(sTempModifiedFilePath));
            if (!bFileCopied)
            {
              string sError = getCatStr("unDeviceTypeCreator","GENERATE-COPYERROR");
              strreplace(sError,"SOURCE",g_sTemplatePath + "/" + baseName(sTempModifiedFilePath));
              strreplace(sError,"DESTINATION",g_sNewPackagePath + "/" + g_sPackageName + "/" + sTempModifiedFilePath);
      
              fwException_raise(
                  dsExceptions,
                  "ERROR",
                  sError,
                  "GENERATE-COPYERROR"
              );
              return -1;
            }
          }
          
          
          int xmlNodeId = xmlAppendChild(g_xmlPackageDescription, g_xmlPackageDescriptionFirstChild, XML_ELEMENT_NODE, XML_NODE_DPLIST);
          xmlAppendChild(g_xmlPackageDescription, xmlNodeId,XML_TEXT_NODE, sTempModifiedFilePath); 
        }
        break;
      }
        
      // ---------------------------------------------------------------- Any other kind of file ---------------------------------------------------------------- \\
      // e.g. message catalog (only element for now)
      // Those files can be overwritten anytime, so the action is the same even when editing an existing package or creating a new one.
      case XML_NODE_FILE:
      {
        // Does this file concern a front-end type, a device type, or none ?
        int iTypeConcerned;
        xmlGetElementAttribute(g_xmlTemplateFileList,nextSibling,XML_NODE_ATTRIBUTE_CONCERNS,iTypeConcerned);

        string sFileContent;
        fileToString(g_sTemplatePath + sFilePath,sFileContent);
        string sTagToReplace;
        switch (iTypeConcerned)
        {
          case CURRENT_TYPE_FRONT_END:
          case CURRENT_TYPE_DEVICE:
          case CURRENT_TYPE_WIDGET:
          {
            mapping mCurrentMapping = _unDeviceTypeCreator_getMappingForType(iTypeConcerned);
            // Copy it as many times as required
            for (int i = 1 ; i <= mappinglen(mCurrentMapping) ; i++)
            {
              string sCurrentTypeName = mappingGetKey(mCurrentMapping,i);  
          
              // Do nothing if the type already exists
              if (_unDeviceTypeCreator_typeExists(sCurrentTypeName,iTypeConcerned,dsExceptions))
              {
                continue;
              }
              if (dynlen(dsExceptions) > 0)
              {
                return -1;
              }
           
          
              string sTempModifiedFilePath = sModifiedFilePath;
              string sTempFileContent = sFileContent;
              _unDeviceTypeCreator_replaceTags(sCurrentTypeName, iTypeConcerned, sTempModifiedFilePath);
              _unDeviceTypeCreator_replaceTags(sCurrentTypeName, iTypeConcerned, sTempFileContent);
              
              
              if (isfile(g_sTemplatePath + DEVICE_TYPE_INFO_FOLDER + sTempModifiedFilePath))
              {
                bool bFileCopied = copyFile(g_sTemplatePath + DEVICE_TYPE_INFO_FOLDER + sTempModifiedFilePath, g_sNewPackagePath + "/" + g_sPackageName + "/" + sTempModifiedFilePath);
                if (!bFileCopied)
                {
                  string sError = getCatStr("unDeviceTypeCreator","GENERATE-COPYERROR");
                  strreplace(sError,"SOURCE",g_sTemplatePath + "/" +DEVICE_TYPE_INFO_FOLDER + sTempModifiedFilePath);
                  strreplace(sError,"DESTINATION",g_sNewPackagePath + "/" + g_sPackageName + "/" + sTempModifiedFilePath);
      
                  fwException_raise(
                      dsExceptions,
                      "ERROR",
                      sError,
                      "GENERATE-COPYERROR"
                  );
                  return -1;
                }            
              } 
              else
              {
                file f = fopen(g_sTemplatePath + "/" + baseName(sTempModifiedFilePath),"w");
                fputs(sTempFileContent,f);
                fclose(f);
              
                bool bFileCopied = copyFile(g_sTemplatePath + "/" + baseName(sTempModifiedFilePath), g_sNewPackagePath + "/" + g_sPackageName + "/" + sTempModifiedFilePath);
                remove(g_sTemplatePath + "/" + baseName(sTempModifiedFilePath));
                if (!bFileCopied)
                {
                  string sError = getCatStr("unDeviceTypeCreator","GENERATE-COPYERROR");
                  strreplace(sError,"SOURCE",g_sTemplatePath + "/" + baseName(sTempModifiedFilePath));
                  strreplace(sError,"DESTINATION",g_sNewPackagePath + "/" + g_sPackageName + "/" + sTempModifiedFilePath);
      
                  fwException_raise(
                      dsExceptions,
                      "ERROR",
                      sError,
                      "GENERATE-COPYERROR"
                  );
                  return -1;
                }
              }
              int xmlNodeIdFile = xmlAppendChild(g_xmlPackageDescription,g_xmlPackageDescriptionFirstChild,XML_ELEMENT_NODE,XML_NODE_FILE);
              xmlAppendChild(g_xmlPackageDescription,xmlNodeIdFile,XML_TEXT_NODE,sTempModifiedFilePath); 
            }
            break; 
          }
            
          // If it concerns no particular type, just copy it once  
          default:
          {
            _unDeviceTypeCreator_replaceTags("", CURRENT_TYPE_NONE, sFileContent);
            
            if (isfile(g_sTemplatePath + DEVICE_TYPE_INFO_FOLDER + sModifiedFilePath))
            {
              bool bFileCopied = copyFile(g_sTemplatePath + DEVICE_TYPE_INFO_FOLDER + sModifiedFilePath, g_sNewPackagePath + "/" + g_sPackageName + "/" + sModifiedFilePath);
              if (!bFileCopied)
              {
                string sError = getCatStr("unDeviceTypeCreator","GENERATE-COPYERROR");
                strreplace(sError,"SOURCE",g_sTemplatePath + "/" +DEVICE_TYPE_INFO_FOLDER + sModifiedFilePath);
                strreplace(sError,"DESTINATION",g_sNewPackagePath + "/" + g_sPackageName + "/" + sModifiedFilePath);
      
                fwException_raise(
                    dsExceptions,
                    "ERROR",
                    sError,
                    "GENERATE-COPYERROR"
                );
                return -1;
              }            
            }  
            else
            {
              file f = fopen(g_sTemplatePath + "/" + baseName(sModifiedFilePath),"w");
              fputs(sFileContent,f);
              fclose(f);
              
              bool bFileCopied = copyFile(g_sTemplatePath + "/" + baseName(sModifiedFilePath), g_sNewPackagePath + "/" + g_sPackageName + "/" + sModifiedFilePath);
              remove(g_sTemplatePath + "/" + baseName(sModifiedFilePath));
              if (!bFileCopied)
              {
                string sError = getCatStr("unDeviceTypeCreator","GENERATE-COPYERROR");
                strreplace(sError,"SOURCE",g_sTemplatePath + "/" + baseName(sModifiedFilePath));
                strreplace(sError,"DESTINATION",g_sNewPackagePath + "/" + g_sPackageName + "/" + sModifiedFilePath);
      
                fwException_raise(
                    dsExceptions,
                    "ERROR",
                    sError,
                    "GENERATE-COPYERROR"
                );
                return -1;
              }
            }
            int xmlNodeIdFile = xmlAppendChild(g_xmlPackageDescription,g_xmlPackageDescriptionFirstChild,XML_ELEMENT_NODE,XML_NODE_FILE);
            xmlAppendChild(g_xmlPackageDescription,xmlNodeIdFile,XML_TEXT_NODE,sModifiedFilePath); 
            break; 
          }       
        }
        
        
        break;
      }
        
      // ---------------------------------------------------------------- Various data to read (non-file) ---------------------------------------------------------------- \\
        
      case XML_NODE_REQUIRED:
      case XML_NODE_REQUIREDINSTALLER:
      case XML_NODE_REQUIREDPVSS:
      {
        int xmlNodeId = xmlAppendChild(g_xmlPackageDescription,g_xmlPackageDescriptionFirstChild,XML_ELEMENT_NODE,xmlNodeName(g_xmlTemplateFileList,nextSibling));
        xmlAppendChild(g_xmlPackageDescription,xmlNodeId,XML_TEXT_NODE,sFilePath); 
        
        break;
      }
    }
    nextSibling = xmlNextSibling(g_xmlTemplateFileList, nextSibling);
  }
   
    
  
  
  // ---------------------------------------------------------------- Process custom 'include' files ---------------------------------------------------------------- \\
  for (int i = 1 ; i <= dynlen(g_ddaDeviceTypeList) ; i++) {
    string deviceType = g_ddaDeviceTypeList[i][LIST_NAME_INDEX];
    if (mappingHasKey(g_mDeviceTypeInfo, deviceType)) {
      dyn_string includes = g_mDeviceTypeInfo[deviceType][DEVICE_TYPE_INDEX_INCLUDE];
      for (int i = 1; i <= dynlen(includes); i++) {
        string filePath = includes[i];
        bool bFileCopied = copyFile(g_sTemplatePath + DEVICE_TYPE_INFO_FOLDER + filePath, g_sNewPackagePath + "/" + g_sPackageName + "/" + filePath);
        if (!bFileCopied)
        {
          string sError = getCatStr("unDeviceTypeCreator","GENERATE-COPYERROR");
          strreplace(sError,"SOURCE",g_sTemplatePath + "/" +DEVICE_TYPE_INFO_FOLDER + filePath);
          strreplace(sError,"DESTINATION",g_sNewPackagePath + "/" + g_sPackageName + "/" + filePath);

          fwException_raise(dsExceptions, "ERROR", sError, "GENERATE-COPYERROR");
          return -1;
        }

        string fileType = XML_NODE_FILE;
        string fileExt = substr(filePath, strlen(filePath) - 3);
        if (fileExt == "pnl") {
          fileType = XML_NODE_PANEL;
        } else if (fileExt == "ctl") {
          fileType = XML_NODE_SCRIPT;
        } else if (fileExt == "ico") {
          fileType = XML_NODE_IMAGE;
        } else if (fileExt == "dpl") {
          fileType = XML_NODE_DPLIST;
        }

        int xmlNodeIdFile = xmlAppendChild(g_xmlPackageDescription, g_xmlPackageDescriptionFirstChild, XML_ELEMENT_NODE, fileType);
        xmlAppendChild(g_xmlPackageDescription, xmlNodeIdFile, XML_TEXT_NODE, filePath);
      }
    }
  }

  
  
  // ------------------------------------------------------------------------------------------------ //
  // ------------------------------------ Generate data import file --------------------------------- //
  // ------------------------------------------------------------------------------------------------ //
  _unDeviceTypeCreator_showProgress("Writing files.","Common files.","Import data.",97);
  
  // ---------------------------------------------------------------- //
  // ----- File data/TAG_FrontEndType_TAG_DeviceType_import.dat ----- //
  // ---------------------------------------------------------------- //  
  
  string  stringFileTAG_FrontEndType_TAG_DeviceType_import_dat;
  fileToString(g_sTemplatePath + "/data/TAG_FrontEndType_TAG_DeviceType_import.dat",stringFileTAG_FrontEndType_TAG_DeviceType_import_dat);
  
  // Set generation date
  strreplace(stringFileTAG_FrontEndType_TAG_DeviceType_import_dat,"yyyy.mm.dd hh:mm:ss",formatTime("%Y.%m.%d %h:%M:%s",getCurrentTime()));
  
  for (int i = 1 ; i <= dynlen(mappingKeys(g_mFrontEndToDeviceMapping)) ; i++)
  {
    for (int j = 1 ; j <= dynlen(g_mFrontEndToDeviceMapping[mappingKeys(g_mFrontEndToDeviceMapping)[i]]) ; j++)
    {
      string tempstringFile = stringFileTAG_FrontEndType_TAG_DeviceType_import_dat;
      file fileTAG_FrontEndType_TAG_DeviceType_import_dat = fopen(g_sTemplatePath + "./" + mappingKeys(g_mFrontEndToDeviceMapping)[i] +"_" + g_mFrontEndToDeviceMapping[mappingKeys(g_mFrontEndToDeviceMapping)[i]][j] + "_import.dat","w");
      
      dyn_string separatedLines = strsplit(stringFileTAG_FrontEndType_TAG_DeviceType_import_dat, "\n");
      
      for (int k = 1 ; k <= dynlen(separatedLines) ; k++)
      {
        if (strpos(separatedLines[k],"#") == 0)
        {
          // If it is a comment, copy the line as-is
          fputs(separatedLines[k] + "\n",fileTAG_FrontEndType_TAG_DeviceType_import_dat);
        }
        else
        {
          // Else replace the tags first        
          string temp = separatedLines[k];
          if (dynlen(g_mDeviceToWidgetMapping[g_mFrontEndToDeviceMapping[mappingKeys(g_mFrontEndToDeviceMapping)[i]][j]]) > 0)
          {
            strreplace(temp,"TAG_WidgetType",g_mDeviceToWidgetMapping[g_mFrontEndToDeviceMapping[mappingKeys(g_mFrontEndToDeviceMapping)[i]][j]][1]);
          }
          else
          {
            strreplace(temp,"TAG_DeviceTypeTAG_WidgetType","");
          }
          strreplace(temp,"TAG_FrontEndType",mappingKeys(g_mFrontEndToDeviceMapping)[i]);
          strreplace(temp,"TAG_DeviceType",g_mFrontEndToDeviceMapping[mappingKeys(g_mFrontEndToDeviceMapping)[i]][j]);  
          fputs(temp + "\n",fileTAG_FrontEndType_TAG_DeviceType_import_dat);
        }
      }      
      int xmlNodeIdFile = xmlAppendChild(g_xmlPackageDescription,g_xmlPackageDescriptionFirstChild,XML_ELEMENT_NODE,XML_NODE_FILE);
      xmlAppendChild(g_xmlPackageDescription,xmlNodeIdFile,XML_TEXT_NODE,"./data/" + mappingKeys(g_mFrontEndToDeviceMapping)[i] +"_" + g_mFrontEndToDeviceMapping[mappingKeys(g_mFrontEndToDeviceMapping)[i]][j] + "_import.dat");
      fclose (fileTAG_FrontEndType_TAG_DeviceType_import_dat);      
  
      // Move the file to the appropriate directory (copy() creates the subdirectories)
      if (!copyFile(
          g_sTemplatePath + "/" + mappingKeys(g_mFrontEndToDeviceMapping)[i] +"_" + g_mFrontEndToDeviceMapping[mappingKeys(g_mFrontEndToDeviceMapping)[i]][j] + "_import.dat",
          g_sNewPackagePath + "/" + g_sPackageName + "/data/" + mappingKeys(g_mFrontEndToDeviceMapping)[i] +"_" + g_mFrontEndToDeviceMapping[mappingKeys(g_mFrontEndToDeviceMapping)[i]][j] + "_import.dat"))
      {
        string sError = getCatStr("unDeviceTypeCreator","GENERATE-COPYERROR");
        strreplace(sError,"SOURCE",g_sTemplatePath + "./" + mappingKeys(g_mFrontEndToDeviceMapping)[i] +"_" + g_mFrontEndToDeviceMapping[mappingKeys(g_mFrontEndToDeviceMapping)[i]][j] + "_import.dat");
        strreplace(sError,"DESTINATION",g_sNewPackagePath + "./data/" + mappingKeys(g_mFrontEndToDeviceMapping)[i] +"_" + g_mFrontEndToDeviceMapping[mappingKeys(g_mFrontEndToDeviceMapping)[i]][j] + "_import.dat");
      
        fwException_raise(
            dsExceptions,
            "ERROR",
            sError,
            "GENERATE-COPYERROR"
        );
        return -1;
      };
      // Delete the old file
      remove(g_sTemplatePath + "./" + mappingKeys(g_mFrontEndToDeviceMapping)[i] +"_" + g_mFrontEndToDeviceMapping[mappingKeys(g_mFrontEndToDeviceMapping)[i]][j] + "_import.dat");
    }
  }
  
  
  // ------------------------------------------------------------------------------------ //
  // ------------------------------------ XML structure --------------------------------- //
  // ------------------------------------------------------------------------------------ //
  
  _unDeviceTypeCreator_showProgress("Writing files.","Common.","XML structure.",98);
	

	// Create document in the template root folder
	xmlDocumentToFile(xmlStructure, g_sTemplatePath + "/_" + g_sPackageName + "_typeConfig.dat");

	// Move the file to the appropriate directory (copy() creates the subdirectories)
	if (!copyFile(g_sTemplatePath + "/_" + g_sPackageName + "_typeConfig.dat",g_sNewPackagePath + "/" + g_sPackageName + "/data/_" + g_sPackageName + "_typeConfig.dat"))
	{
		string sError = getCatStr("unDeviceTypeCreator","GENERATE-COPYERROR");
		strreplace(sError,"SOURCE",g_sTemplatePath + "/" + g_sPackageName + "_typeConfig.dat");
		strreplace(sError,"DESTINATION",g_sNewPackagePath + "/" + g_sPackageName + "/data/_" + g_sPackageName + "_typeConfig.dat");
			
		fwException_raise(
				dsExceptions,
				"ERROR",
				sError,
				"GENERATE-COPYERROR"
		);

		return -1;
	};
	// Delete the old file
	remove(g_sTemplatePath + "/_" + g_sPackageName + "_typeConfig.dat");
				
	// Add to the XML component file
	int xmlNodeIdFile = xmlAppendChild(g_xmlPackageDescription,g_xmlPackageDescriptionFirstChild,XML_ELEMENT_NODE,XML_NODE_FILE);
	xmlAppendChild(g_xmlPackageDescription,xmlNodeIdFile,XML_TEXT_NODE,"./data/_" + g_sPackageName + "_typeConfig.dat");
 
  
  
  
  // ------------------------------------------------------------------------------------ //
  // ------------------------------------ Config File ----------------------------------- //
  // ------------------------------------------------------------------------------------ //
  
  _unDeviceTypeCreator_showProgress("Writing files.","Config file.","",99);
  // Create document in the template root folder
  file fConfig = fopen(g_sTemplatePath + "/" + g_sPackageName + ".config","w");

  fputs("[ui]\n",fConfig);  
  for (int i = 1 ; i <= dynlen(g_dsLibListUiOnly) ; i++)
  {
    fputs("LoadCtrlLibs = \"" + g_dsLibListUiOnly[i] + "\"\n",fConfig);  
  }
  for (int i = 1 ; i <= dynlen(g_dsLibList) ; i++)
  {
    fputs("LoadCtrlLibs = \"" + g_dsLibList[i] + "\"\n",fConfig);  
  }

  fputs("[ctrl]\n",fConfig); 
  for (int i = 1 ; i <= dynlen(g_dsLibList) ; i++)
  {
    fputs("LoadCtrlLibs = \"" + g_dsLibList[i] + "\"\n",fConfig);  
  } 
  
  fclose(fConfig);
  
  if (!copyFile(g_sTemplatePath + "/" + g_sPackageName + ".config",g_sNewPackagePath + "/" + g_sPackageName + "/config/" + g_sPackageName + ".config","w"))
  {
    // Delete the old file
    remove(g_sTemplatePath + "/" + g_sPackageName + ".config");
    string sError = getCatStr("unDeviceTypeCreator","GENERATE-COPYERROR");
    strreplace(sError,"SOURCE",g_sTemplatePath + "/" + g_sPackageName + ".config");
    strreplace(sError,"DESTINATION",g_sNewPackagePath + "/" + g_sPackageName + "/config/" + g_sPackageName + ".config");
      
    fwException_raise(
        dsExceptions,
        "ERROR",
        sError,
        "GENERATE-COPYERROR"
    );
    return -1;
  }
  // Delete the old file
  remove(g_sTemplatePath + "/" + g_sPackageName + ".config");
  
  int xmlNodeIdFile = xmlAppendChild(g_xmlPackageDescription,g_xmlPackageDescriptionFirstChild,XML_ELEMENT_NODE,XML_NODE_CONFIG);
  xmlAppendChild(g_xmlPackageDescription,xmlNodeIdFile,XML_TEXT_NODE,"./config/" + g_sPackageName + ".config"); 
  
  
  
  
  // ------------------------------------------------------------------------------------ //
  // ------------------------------------ XML component --------------------------------- //
  // ------------------------------------------------------------------------------------ //
  // This is the XML file used by fwInstallation to install the package
   
  _unDeviceTypeCreator_showProgress("Writing files.","Component.","",100);
  // Create document in the template root folder
  xmlDocumentToFile(g_xmlPackageDescription,g_sTemplatePath + "/" + g_sPackageName + ".xml");
  
  // Move the file to the appropriate directory (copy() creates the subdirectories)
  
  if (!copyFile(g_sTemplatePath + "/" + g_sPackageName + ".xml",g_sNewPackagePath + "/" + g_sPackageName + "/" + g_sPackageName + ".xml"))
  {
    // Delete the old file
    remove(g_sTemplatePath + "/" + g_sPackageName + ".xml");
    string sError = getCatStr("unDeviceTypeCreator","GENERATE-COPYERROR");
    strreplace(sError,"SOURCE",g_sTemplatePath + "/" + g_sPackageName + ".xml");
    strreplace(sError,"DESTINATION",g_sNewPackagePath + "/" + g_sPackageName + "/" + g_sPackageName + ".xml");
      
    fwException_raise(
        dsExceptions,
        "ERROR",
        sError,
        "GENERATE-COPYERROR"
    );
    return -1;
  }
  // Delete the old file
  remove(g_sTemplatePath + "/" + g_sPackageName + ".xml");
  
 
 
 
	return 0;
}

//_unDeviceTypeCreator_createNewTypeScript
/**
  @par Description:
  Creates a new CTRL script from template.
  
  @par Usage: 
  Internal
  
	
	@param sTypeName          input, the type concerned by this script.
	@param iTypeType          input, indicates wether it is a front-end (CURRENT_TYPE_FRONT_END), a device (CURRENT_TYPE_DEVICE), a widget type (CURRENT_TYPE_WIDGET) or none (CURRENT_TYPE_NONE, results in an error).
	@param xmlDocumentId      input, the XML template file to generate the script from.
	@param xmlRootNodeId      input, the root node of this XML file
	@param dsExceptions      output, list of errors that occurred during the process.
	@param dsExistingFilePath input, if the file already exists, indicates its path.
*/
string _unDeviceTypeCreator_createNewTypeScript(string sTypeName, int iType, unsigned xmlDocumentId, unsigned xmlRootNodeId, dyn_string &dsExceptions, string dsExistingFilePath)
{
  string sNewFileContent;

  // Do some lines have to be duplicated for every device linked to this front-end ?
  mapping mXmlAttributes = xmlElementAttributes(xmlDocumentId,xmlRootNodeId);
  bool bNeedToDuplicate = false;
  bool bBrowseSubTypes = false;
  bool bBrowseSuperTypes = false;
  dyn_string sTypesToBrowse = makeDynString();
  mapping mRelevantMapping = _unDeviceTypeCreator_getMappingForType(iType);
  if (mappingHasKey(mXmlAttributes,XML_TEMPLATE_CTRLSCRIPT_ATTR_SUBTYPES))
  {
    bBrowseSubTypes = mXmlAttributes[XML_TEMPLATE_CTRLSCRIPT_ATTR_SUBTYPES];
    sTypesToBrowse = mRelevantMapping[sTypeName];
  }
  else if (mappingHasKey(mXmlAttributes,XML_TEMPLATE_CTRLSCRIPT_ATTR_SUPERTYPES))
  {
    bBrowseSuperTypes = mXmlAttributes[XML_TEMPLATE_CTRLSCRIPT_ATTR_SUPERTYPES];
    sTypesToBrowse = _unDeviceTypeCreator_getSuperTypes(iType,sTypeName);
  }
  bNeedToDuplicate = bBrowseSuperTypes || bBrowseSubTypes;  
  
  // Get file header content
  unsigned xmlFileHeaderNodeId = xmlFirstChild(xmlDocumentId,xmlRootNodeId);
  string sNewFileHeader = xmlNodeValue(xmlDocumentId, xmlFirstChild(xmlDocumentId,xmlFileHeaderNodeId));
  _unDeviceTypeCreator_replaceTags(sTypeName, iType, sNewFileHeader);
  
  sNewFileContent += sNewFileHeader + "\n";
  sNewFileContent += DOXYGEN_START_TAG + "\n\n";
  

  if (bUpdatingExistingPackage)
  {
    // Get the name of the file that will be created.
    string sFileName;
    xmlGetElementAttribute(xmlDocumentId, xmlRootNodeId, XML_TEMPLATE_CTRLSCRIPT_ATTR_FILENAME, sFileName);
    _unDeviceTypeCreator_replaceTags(sTypeName, iType, sFileName);  
    // If this file already exists, its content is appended to the new file content.
    if (_unDeviceTypeCreator_fileExists(dsExistingFilePath))
    {
      string sFormerFileContent;
      fileToString(mLoadedPackageLocation + "/" + dsExistingFilePath, sFormerFileContent);
      // Extract content between doxygen tags
      sFormerFileContent = substr(
          sFormerFileContent,
          strpos(sFormerFileContent,DOXYGEN_START_TAG) + strlen(DOXYGEN_START_TAG),
          strpos(sFormerFileContent,DOXYGEN_END_TAG) - (strpos(sFormerFileContent,DOXYGEN_START_TAG) + strlen(DOXYGEN_START_TAG)));
      strrtrim(sFormerFileContent);
      strltrim(sFormerFileContent);
      sNewFileContent += sFormerFileContent;
    }
  }
  
 
  // Read the element list
  unsigned xmlElementListNodeId = xmlNextSibling(xmlDocumentId, xmlFileHeaderNodeId);
      
  unsigned nextSibling = xmlFirstChild(xmlDocumentId,xmlElementListNodeId);
  while (nextSibling != -1)
  {
    
    switch (xmlNodeName(xmlDocumentId,nextSibling))
    {
      
      // -------------------------------- Append a constant -------------------------------- \\
      
      case XML_TEMPLATE_CONSTANT:
      {
        // ID of the constant
        string sConstantName;
        xmlGetElementAttribute(xmlDocumentId,nextSibling,XML_TEMPLATE_CONSTANT_ATTR_NAME,sConstantName);
        _unDeviceTypeCreator_replaceTags(sTypeName, iType, sConstantName);    
        
        // Does this constant need to be duplicated ?
        bool bDuplicate = false;
        xmlGetElementAttribute(xmlDocumentId,nextSibling,XML_TEMPLATE_ATTR_DUPLICATE_ELEMENT,bDuplicate);
        
        // Does this constant comment need to be duplicated ?
        bool bDuplicateComment;
        xmlGetElementAttribute(xmlDocumentId,nextSibling,XML_TEMPLATE_ATTR_DUPLICATE_COMMENT,bDuplicateComment);
        
        // Get comment and content nodes
        unsigned xmlConstantCommentNodeId;
        string sConstantComment;
        // Does the comment have to be left as-is ?
        bool bNoReplace = false;
        unsigned xmlConstantContentNodeId;
        string sConstantContent;
        xmlConstantCommentNodeId = xmlFirstChild(xmlDocumentId,nextSibling);
        
        // There can sometimes be no comment, in this case ("else" block) the first child is the content
        if (xmlNodeName(xmlDocumentId,xmlConstantCommentNodeId) == XML_TEMPLATE_CONSTANT_COMMENT)
        {
          xmlGetElementAttribute(xmlDocumentId,xmlConstantCommentNodeId,XML_TEMPLATE_CONSTANT_ATTR_NOREPLACE,bNoReplace);
          
          // Comment value
          sConstantComment = xmlNodeValue(xmlDocumentId,xmlFirstChild(xmlDocumentId,xmlConstantCommentNodeId));
          
          if (!bNoReplace)
          {
            _unDeviceTypeCreator_replaceTags(sTypeName, iType, sConstantComment);     
          }
          // If we don't want to replace anything, then we don't need to duplicate anything.
          else
          {             
            bDuplicateComment = false; 
          }
          
          xmlConstantContentNodeId = xmlNextSibling(xmlDocumentId,xmlConstantCommentNodeId);
        }
        else
        {
          xmlConstantContentNodeId = xmlConstantCommentNodeId;
        }
        
        // Content value
        // Warning, content can be empty. In this case the firstChild will be -1 and trying to access a nodeValue() would generate an error.
        sConstantContent = xmlFirstChild(xmlDocumentId,xmlConstantContentNodeId) == -1 ? "" : xmlNodeValue(xmlDocumentId,xmlFirstChild(xmlDocumentId,xmlConstantContentNodeId));
        _unDeviceTypeCreator_replaceTags(sTypeName, iType, sConstantContent);    
        

        string sTempConstant;
          
        if (bDuplicate && bNeedToDuplicate)
        {
          bool bThereIsANewConstant = false;
          for (int i = 1 ; i <= dynlen(sTypesToBrowse) ; i++)
          {
            string sTempConstantName = sConstantName;
            _unDeviceTypeCreator_replaceTags(sTypesToBrowse[i],bBrowseSubTypes? iType-1 : iType+1, sTempConstantName);
            if (bUpdatingExistingPackage && globalExists(sTempConstantName) && (sConstantName != ""))
            {
              continue;
            }
            bThereIsANewConstant = true;
            // If the constant is not defined yet it can be appended to the file.
                    
            if (bDuplicateComment)
            {
              // Copy of the original string to keep the tags.
              string sTempStringComment = sConstantComment;         
              if (!bNoReplace)
              {
                _unDeviceTypeCreator_replaceTags(sTypesToBrowse[i],bBrowseSubTypes? iType-1 : iType+1,sTempStringComment);
              }
              sTempConstant += sTempStringComment + "\n";
            }
            // Copy of the original string to keep the tags.
            string sTempStringContent = sConstantContent;
            _unDeviceTypeCreator_replaceTags(sTypesToBrowse[i],bBrowseSubTypes? iType-1 : iType+1,sTempStringContent);
            sTempConstant += sTempStringContent + "\n";
          }
          
          
          if (!bDuplicateComment && bThereIsANewConstant)
          {
              sTempConstant = sConstantComment + sTempConstant + "\n";
          }
          
        }
        else
        {  
          if (bUpdatingExistingPackage && globalExists(sConstantName) && (sConstantName != ""))
          {
            break;
          }
          
          sTempConstant += sConstantComment + "\n";
          sTempConstant += sConstantContent + "\n";
        }
        
        
        sNewFileContent += sTempConstant;    
        
        
        
        
        break;
      }
      
      // -------------------------------- Append a function -------------------------------- \\
      
      case XML_TEMPLATE_FUNCTION:
      {
        // Does this function need to be duplicated ?
        bool bDuplicate = false;
        xmlGetElementAttribute(xmlDocumentId,nextSibling,XML_TEMPLATE_ATTR_DUPLICATE_ELEMENT,bDuplicate);
        
        // Get the name of the function
        string sFunctionName;
        xmlGetElementAttribute(xmlDocumentId,nextSibling,XML_TEMPLATE_FUNCTION_ATTR_NAME,sFunctionName);
				
				// Ignore function?
				// Note: no "break 2;" possible in a CTRL script
				bool bIgnore = false;
				if ((CURRENT_TYPE_DEVICE == iType) && (mappingHasKey(g_mDeviceTypeInfo, sTypeName)))
				{
					for (int i = 1 ; i <= dynlen(g_mDeviceTypeInfo[sTypeName][DEVICE_TYPE_INDEX_IGNORE_FUNCTIONS]) ; i++)
					{
						if (_unDeviceTypeCreator_stringEndsBy(sFunctionName, g_mDeviceTypeInfo[sTypeName][DEVICE_TYPE_INDEX_IGNORE_FUNCTIONS][i]))
						{
							bIgnore = true;
						}
					}
				}
				
				if (bIgnore)
				{
					break;
				}
				
				
        _unDeviceTypeCreator_replaceTags(sTypeName,iType,sFunctionName);
        
        // Function header
        unsigned xmlFunctionHeaderNodeId = xmlFirstChild(xmlDocumentId,nextSibling);
        string sFunctionHeader = xmlNodeValue(xmlDocumentId,xmlFirstChild(xmlDocumentId,xmlFunctionHeaderNodeId));
        _unDeviceTypeCreator_replaceTags(sTypeName,iType,sFunctionHeader);
        
        // Function content
        unsigned xmlFunctionContentNodeId = xmlNextSibling(xmlDocumentId,xmlFunctionHeaderNodeId);
        string sFunctionContent = xmlNodeValue(xmlDocumentId,xmlFirstChild(xmlDocumentId,xmlFunctionContentNodeId));
        _unDeviceTypeCreator_replaceTags(sTypeName,iType,sFunctionContent);
        
        
        if (bDuplicate)
        {
          for (int i = 1 ; i <= dynlen(sTypesToBrowse) ; i++)
          {
        
            string sTempFunctionName = sFunctionName;
            string sTempHeader = sFunctionHeader;
            string sTempContent = sFunctionContent;
            _unDeviceTypeCreator_replaceTags(sTypesToBrowse[i],bBrowseSubTypes? iType-1 : iType+1,sTempHeader);
            _unDeviceTypeCreator_replaceTags(sTypesToBrowse[i],bBrowseSubTypes? iType-1 : iType+1,sTempContent);
            _unDeviceTypeCreator_replaceTags(sTypesToBrowse[i],bBrowseSubTypes? iType-1 : iType+1,sTempFunctionName);
            
            // Check if the function has already been defined
            // If not we append it to the file. Else we do nothing.
            if ((bUpdatingExistingPackage && !isFunctionDefined(sTempFunctionName)) || !bUpdatingExistingPackage)
            {
              sNewFileContent += sTempHeader;
              sNewFileContent += sTempContent + "\n";  
            }
          } 
        }
        else
        {
            if ((bUpdatingExistingPackage && !isFunctionDefined(sFunctionName)) || !bUpdatingExistingPackage)
            {
              sNewFileContent += sFunctionHeader;
              sNewFileContent += sFunctionContent + "\n";    
            }   
        }
        
        break;      
      }
      
      
      
      // Cannot be anything else than constant or function (XSD rules have been validated).
      
    }
    
    nextSibling = xmlNextSibling(xmlDocumentId,nextSibling);
  }
  
  
  
  
  
  sNewFileContent += "\n" + DOXYGEN_END_TAG;
  
  
  // Replace all special device type information (if the script is for a device)
  if (CURRENT_TYPE_DEVICE == iType)
  {
    if (mappingHasKey(g_mDeviceTypeInfo, sTypeName))
    {
      for (int i = 1 ; i <= mappinglen(g_mDeviceTypeInfo[sTypeName][DEVICE_TYPE_INDEX_SCRIPT]) ; i++)
      {
        string sTag = mappingGetKey(g_mDeviceTypeInfo[sTypeName][DEVICE_TYPE_INDEX_SCRIPT], i);
        string sContent = mappingGetValue(g_mDeviceTypeInfo[sTypeName][DEVICE_TYPE_INDEX_SCRIPT], i);
        
        strreplace(sContent, "\r", "");
        sContent = strrtrim(sContent);
        sContent = strltrim(sContent, "\n");
        strreplace(sNewFileContent, sTag, sContent);
      }
    }
  }
  
  return sNewFileContent;
}

//_unDeviceTypeCreator_exportStructure
/**
	@par Descriptions:
	Function to export the structure of the package.
	This saves the mapping, front-end, device and widget informations to an XML file
	This XML is used by the function importStructurem which loads the XML file to rebuild the mapping, front-end, device and widget informations.
	The parameters passed to the function are the 4 text fields of the interface. When the XML file is loaded, thos fields are automatically filled.
	
	Structure (example data) :
  <package name="packageName" prefix="prefix">
  
    <lists>
      <frontEndList>
        <frontEndType exists="true" name="frontEnd1"/>
        <frontEndType exists="false" name="frontEnd2"/>
        ...
      <frontEndList> 
      
      <deviceList>
        <deviceType exists="true" name="device1"/>
        <deviceType exists="false" name="device2"/>
        ...
      </deviceList>
      
      <widgetList>
        <widgetType exists="true" name="widget1"/>
        <widgetType exists="false" name="widget2"/>
        ...      
      </widgetList>      
    </lists>
    
    
    <mapping>
    
      <frontEndDevice>
        <frontEndType name="frontEnd1">
          <deviceType name="device1"/>
          <deviceType name="device2"/>
        </frontEndType>        
        <frontEndType name="frontEnd2">
          <deviceType name="device1"/>
          <deviceType name="device3"/>
        </frontEndType>
        ...    
      </frontEndDevice>
      
      <deviceWidget>
        <deviceType name="device1">
          <widgetType name="widget1"/>
          <widgetType name="widget2"/>
        </deviceType>
        <deviceType name="device2">
          <widgetType name="widget5"/>
          <widgetType name="widget6"/>
        </deviceType>
      </deviceWidget>
      
    </mapping>
  </package>
	
	
	@param packageName input, the name of the package
	@param prefix 		 input, the prefix used in this package
	
	@return the created XML file.

*/
unsigned _unDeviceTypeCreator_exportStructure(string packageName, string prefix)
{
  unsigned outputXmlFile = xmlNewDocument();
  
  // Root tag (package)
  int xmlNodeIdPackage = xmlAppendChild(outputXmlFile,-1,XML_ELEMENT_NODE,XML_IMPORT_EXPORT_NODE_PACKAGE);
  // Attribute package name
  xmlSetElementAttribute(outputXmlFile, xmlNodeIdPackage, XML_IMPORT_EXPORT_ATTR_NAME, packageName);
  // Attribute package prefix
  xmlSetElementAttribute(outputXmlFile, xmlNodeIdPackage, XML_IMPORT_EXPORT_ATTR_PREFIX, prefix);
  
  // Type lists
  int xmlNodeIdLists = xmlAppendChild(outputXmlFile,xmlNodeIdPackage,XML_ELEMENT_NODE,XML_IMPORT_EXPORT_NODE_LISTS);
  
  // Front-end type list
  int xmlNodeIdListFrontEnd = xmlAppendChild(outputXmlFile,xmlNodeIdLists,XML_ELEMENT_NODE,XML_IMPORT_EXPORT_NODE_FRONTENDLIST);
  
	for (int i = 1 ; i <= dynlen(g_ddaFrontEndTypeList) ; i++)
  {
      string currentFrontEndTypeName = g_ddaFrontEndTypeList[i][LIST_NAME_INDEX];
      bool currentFrontEndTypeExists = g_ddaFrontEndTypeList[i][LIST_EXISTS_INDEX];    
      
      int xmlNodeIdListFronEndElt = xmlAppendChild(outputXmlFile,xmlNodeIdListFrontEnd,XML_ELEMENT_NODE,XML_IMPORT_EXPORT_NODE_FRONTENDTYPE);
      xmlSetElementAttribute(outputXmlFile, xmlNodeIdListFronEndElt, XML_IMPORT_EXPORT_ATTR_NAME, currentFrontEndTypeName);      
      xmlSetElementAttribute(outputXmlFile, xmlNodeIdListFronEndElt, XML_IMPORT_EXPORT_ATTR_EXISTS, currentFrontEndTypeExists);
      
  }
  
  // Device type list
  int xmlNodeIdListDevice = xmlAppendChild(outputXmlFile,xmlNodeIdLists,XML_ELEMENT_NODE,XML_IMPORT_EXPORT_NODE_DEVICELIST);
  
	for (int i = 1 ; i <= dynlen(g_ddaDeviceTypeList) ; i++)
  {
      string currentDeviceName = g_ddaDeviceTypeList[i][LIST_NAME_INDEX];
      bool currentDeviceExists = g_ddaDeviceTypeList[i][LIST_EXISTS_INDEX];     
      
      int xmlNodeIdListDeviceElt = xmlAppendChild(outputXmlFile,xmlNodeIdListDevice,XML_ELEMENT_NODE,XML_IMPORT_EXPORT_NODE_DEVICETYPE);
      xmlSetElementAttribute(outputXmlFile, xmlNodeIdListDeviceElt, XML_IMPORT_EXPORT_ATTR_NAME, currentDeviceName);      
      xmlSetElementAttribute(outputXmlFile, xmlNodeIdListDeviceElt, XML_IMPORT_EXPORT_ATTR_EXISTS, currentDeviceExists);
      
  }  
  
  // Widget type list
  int xmlNodeIdListWidget = xmlAppendChild(outputXmlFile,xmlNodeIdLists,XML_ELEMENT_NODE,XML_IMPORT_EXPORT_NODE_WIDGETLIST);
  
	for (int i = 1 ; i <= dynlen(g_ddaWidgetTypeList) ; i++)
  {
      string currentWidgetName = g_ddaWidgetTypeList[i][LIST_NAME_INDEX];
			if ("" == currentWidgetName)
			{
				currentWidgetName = EMPTY_WIDGET_IDENTIFIER;
			}
      bool currentWidgetExists = g_ddaWidgetTypeList[i][LIST_EXISTS_INDEX];   
      
      int xmlNodeIdListWidgetElt = xmlAppendChild(outputXmlFile,xmlNodeIdListWidget,XML_ELEMENT_NODE,XML_IMPORT_EXPORT_NODE_WIDGETTYPE);
      xmlSetElementAttribute(outputXmlFile, xmlNodeIdListWidgetElt, XML_IMPORT_EXPORT_ATTR_NAME, currentWidgetName);      
      xmlSetElementAttribute(outputXmlFile, xmlNodeIdListWidgetElt, XML_IMPORT_EXPORT_ATTR_EXISTS, currentWidgetExists);
      
  }
  
  
  
  
  
  // Mapping
  int xmlNodeIdMapping = xmlAppendChild(outputXmlFile,xmlNodeIdPackage,XML_ELEMENT_NODE,XML_IMPORT_EXPORT_NODE_MAPPING);
  
  // Front-end/device mapping
  int xmlNodeIdMappingFrontEndDevice = xmlAppendChild(outputXmlFile,xmlNodeIdMapping,XML_ELEMENT_NODE,XML_IMPORT_EXPORT_NODE_FRONTENDDEVICE);
  for (int i = 1 ; i <= mappinglen(g_mFrontEndToDeviceMapping) ; i++)
  {
    int xmlNodeIdMappingFrontEndDeviceFrontEndi = xmlAppendChild(outputXmlFile, xmlNodeIdMappingFrontEndDevice, XML_ELEMENT_NODE, XML_IMPORT_EXPORT_NODE_FRONTENDTYPE);
    xmlSetElementAttribute(outputXmlFile, xmlNodeIdMappingFrontEndDeviceFrontEndi, XML_IMPORT_EXPORT_ATTR_NAME, mappingGetKey(g_mFrontEndToDeviceMapping,i));
    
    for (int j = 1 ; j <= dynlen(mappingGetValue(g_mFrontEndToDeviceMapping,i)) ; j++)
    {
      int xmlNodeIdMappingFrontEndDeviceFrontEndiDevicej = xmlAppendChild(outputXmlFile,xmlNodeIdMappingFrontEndDeviceFrontEndi,XML_ELEMENT_NODE,XML_IMPORT_EXPORT_NODE_DEVICETYPE);
      xmlSetElementAttribute(outputXmlFile, xmlNodeIdMappingFrontEndDeviceFrontEndiDevicej, XML_IMPORT_EXPORT_ATTR_NAME, mappingGetValue(g_mFrontEndToDeviceMapping,i)[j]);
    }
  }
  
  // Device/widget mapping
  int xmlNodeIdMappingDeviceWidget = xmlAppendChild(outputXmlFile, xmlNodeIdMapping, XML_ELEMENT_NODE, XML_IMPORT_EXPORT_NODE_DEVICEWIDGET);
  for (int i = 1 ; i <= mappinglen(g_mDeviceToWidgetMapping) ; i++)
  {
		
    int xmlNodeIdMappingDeviceWidgetDevicei = xmlAppendChild(outputXmlFile, xmlNodeIdMappingDeviceWidget, XML_ELEMENT_NODE, XML_IMPORT_EXPORT_NODE_DEVICETYPE);
    xmlSetElementAttribute(outputXmlFile, xmlNodeIdMappingDeviceWidgetDevicei, XML_IMPORT_EXPORT_ATTR_NAME, mappingGetKey(g_mDeviceToWidgetMapping,i));
    
    for (int j = 1 ; j <= dynlen(mappingGetValue(g_mDeviceToWidgetMapping,i)) ; j++)
    {
			string sWidget = mappingGetValue(g_mDeviceToWidgetMapping,i)[j];
			if ("" == sWidget)
			{
				sWidget = EMPTY_WIDGET_IDENTIFIER;
			}
			
      int xmlNodeIdMappingDeviceWidgetDeviceiWidgetj = xmlAppendChild(outputXmlFile, xmlNodeIdMappingDeviceWidgetDevicei, XML_ELEMENT_NODE, XML_IMPORT_EXPORT_NODE_WIDGETTYPE);
      xmlSetElementAttribute(outputXmlFile, xmlNodeIdMappingDeviceWidgetDeviceiWidgetj, XML_IMPORT_EXPORT_ATTR_NAME, sWidget);
    }
  }
  
  
  
  return outputXmlFile;  
}

//_unDeviceTypeCreator_importStructure
/**
  @par Description:
  Load a package previously generated with _unDeviceTypeCreator_exportStructure.
  
  @par Usage: 
  Internal
  
	@param sFilePath     input, path to the file to load data from.
	@param dsExceptions output, list of errors that occurred during the process.
*/
void _unDeviceTypeCreator_importStructure(string sFilePath, dyn_string &dsExceptions)
{
  sLoadPackacteStructureFrom = sFilePath;
  // Open XML file
  string errMsg;
  int errLine;
  int errColumn;  
  int packageInfoXml = xmlDocumentFromFile(sFilePath,errMsg,errLine,errColumn);
  // Syntax error ?
  if (packageInfoXml == -1)
  {
    string sError = getCatStr("unDeviceTypeCreator","LOAD-XMLSYNTAXERROR");
    strreplace(sError,"ERRMSG",errMsg);
    strreplace(sError,"LINENUM",errLine);
    strreplace(sError,"COLUMNUM",errColumn);
      
    fwException_raise(
        dsExceptions,
        "ERROR",
        sError,
        "LOAD-XMLSYNTAXERROR"
    ); 
    return; 
  }
  
  
  string sPackageName;
  string sPrefix;
  
  
   
  // Access "package" tag
  int xmlPackageFirstChild = xmlFirstChild(packageInfoXml);
  // First tag of the file is not "package" ? Generate error
  if (xmlNodeName(packageInfoXml,xmlPackageFirstChild) != XML_IMPORT_EXPORT_NODE_PACKAGE)
  {
    string sError = getCatStr("unDeviceTypeCreator","LOAD-INCORRECTXMLSTRUCTURE");
    strreplace(sError,"TAG",xmlNodeName(packageInfoXml,xmlPackageFirstChild));
      
    fwException_raise(
        dsExceptions,
        "ERROR",
        sError,
        "LOAD-INCORRECTXMLSTRUCTURE"
    ); 
    return; 
  }
  // Read package name;
  xmlGetElementAttribute(packageInfoXml,xmlPackageFirstChild,XML_IMPORT_EXPORT_ATTR_NAME,sPackageName);
  if (sPackageName == "")
  {
    string sError = getCatStr("unDeviceTypeCreator","LOAD-ATTRIBUTEEMPTYERROR");
    strreplace(sError,"ATTR",XML_IMPORT_EXPORT_ATTR_NAME);
    strreplace(sError,"NODE",xmlNodeName(packageInfoXml,xmlPackageFirstChild));
    
    fwException_raise(
        dsExceptions,
        "ERROR",
        sError,
        "LOAD-ATTRIBUTEEMPTYERROR"
    ); 
    return; 
  }  
  // Read package name;
  xmlGetElementAttribute(packageInfoXml,xmlPackageFirstChild,XML_IMPORT_EXPORT_ATTR_PREFIX,sPrefix);
  if (sPrefix == "")
  {
    string sError = getCatStr("unDeviceTypeCreator","LOAD-ATTRIBUTEEMPTYERROR");
    strreplace(sError,"ATTR",XML_IMPORT_EXPORT_ATTR_NAME);
    strreplace(sError,"NODE",xmlNodeName(packageInfoXml,xmlPackageFirstChild));
    
    fwException_raise(
        dsExceptions,
        "ERROR",
        sError,
        "LOAD-ATTRIBUTEEMPTYERROR"
    ); 
    return; 
  }  
  
  
  
  // Access first child inside "package"
  int xmlPackageFirstChild = xmlFirstChild(packageInfoXml,xmlPackageFirstChild);
  
  // XML structure error ?
  if (xmlPackageFirstChild == -1)
  {
    string sError = getCatStr("unDeviceTypeCreator","LOAD-INCORRECTXMLSTRUCTURE2");
      
    fwException_raise(
        dsExceptions,
        "ERROR",
        sError,
        "LOAD-INCORRECTXMLSTRUCTURE2"
    ); 
    return; 
  }
 


  // Temporary mapping variable. If the loading succeeds then the real variable is replaced.
  mapping localFrontEndDeviceMapping;
  mapping localDeviceWidgetMapping;
  dyn_dyn_anytype localFrontEndList;
  dyn_dyn_anytype localDeviceList;
  dyn_dyn_anytype localWidgetList;
  
  // Iterate through each tag of the XML file
  unsigned nextSibling = xmlPackageFirstChild;
  while (nextSibling != -1)
  {
    switch (xmlNodeName(packageInfoXml,nextSibling))
    {
      
      // Read type lists
      case XML_IMPORT_EXPORT_NODE_LISTS:
      {
        unsigned xmlListsFirstChildNextSibling = xmlFirstChild(packageInfoXml,nextSibling);
        while (xmlListsFirstChildNextSibling != -1)
        {
          switch (xmlNodeName(packageInfoXml,xmlListsFirstChildNextSibling))
          {       
            // ---------------------------------------------------------------- //
            // -------------------- Front-end type list ----------------------- //
            // ---------------------------------------------------------------- //
            case XML_IMPORT_EXPORT_NODE_FRONTENDLIST:
            { 
              unsigned xmlFrontEndTypeListChildNextSibling = xmlFirstChild(packageInfoXml,xmlListsFirstChildNextSibling);
              while (xmlFrontEndTypeListChildNextSibling != -1)
              {
                // Child tag is not "frontEndType" ? Generate error
                if (xmlNodeName(packageInfoXml,xmlFrontEndTypeListChildNextSibling) != XML_IMPORT_EXPORT_NODE_FRONTENDTYPE)
                {
                  string sError = getCatStr("unDeviceTypeCreator","LOAD-INCORRECTXMLSTRUCTURE");
                  strreplace(sError,"TAG",xmlNodeName(packageInfoXml,xmlFrontEndTypeListChildNextSibling));
      
                  fwException_raise(
                      dsExceptions,
                      "ERROR",
                      sError,
                      "LOAD-INCORRECTXMLSTRUCTURE"
                  ); 
                  return; 
                }
                
                // Read front-end type name
                string sFrontEndTypeName;
                xmlGetElementAttribute(packageInfoXml,xmlFrontEndTypeListChildNextSibling,XML_IMPORT_EXPORT_ATTR_NAME,sFrontEndTypeName);
                if (sFrontEndTypeName == "")
                {
                  string sError = getCatStr("unDeviceTypeCreator","LOAD-ATTRIBUTEEMPTYERROR");
                  strreplace(sError,"ATTR",XML_IMPORT_EXPORT_ATTR_NAME);
                  strreplace(sError,"NODE",xmlNodeName(packageInfoXml,xmlFrontEndTypeListChildNextSibling));
    
                  fwException_raise(
                      dsExceptions,
                      "ERROR",
                      sError,
                      "LOAD-ATTRIBUTEEMPTYERROR"
                  ); 
                  return; 
                }  
                
                // Read front-end type exists (as a string "true"/"false")
                string sFrontEndTypeExists;
                xmlGetElementAttribute(packageInfoXml,xmlFrontEndTypeListChildNextSibling,XML_IMPORT_EXPORT_ATTR_EXISTS,sFrontEndTypeExists);
                // Convert to bool
                bool bFrontEndTypeExists;
                switch (sFrontEndTypeExists)
                {
                  case "TRUE":
                  {
                    bFrontEndTypeExists = true;
                    break;                  
                  }
                  case "FALSE":
                  {
                    bFrontEndTypeExists = false;
                    break;                  
                  }
                  
                  default:
                  {
                    string sError = getCatStr("unDeviceTypeCreator","LOAD-INVALIDXMLBOOL");
                    strreplace(sError,"VALUE",sFrontEndTypeExists);
                    strreplace(sError,"NODE",xmlNodeName(packageInfoXml,xmlFrontEndTypeListChildNextSibling));
    
                    fwException_raise(
                        dsExceptions,
                        "ERROR",
                        sError,
                        "LOAD-INVALIDXMLBOOL"
                    ); 
                    return; 
                  }
                }
                dynAppend(localFrontEndList,makeDynAnytype(sFrontEndTypeName,bFrontEndTypeExists));
                xmlFrontEndTypeListChildNextSibling = xmlNextSibling(packageInfoXml, xmlFrontEndTypeListChildNextSibling);
              }
              break;
            }
            // Device type list
            case XML_IMPORT_EXPORT_NODE_DEVICELIST:
            {       
              // ---------------------------------------------------------------- //
              // ---------------------- Device type list ------------------------ //
              // ---------------------------------------------------------------- //
              unsigned xmlDeviceTypeListChildNextSibling = xmlFirstChild(packageInfoXml,xmlListsFirstChildNextSibling);
              while (xmlDeviceTypeListChildNextSibling != -1)
              {
                // Child tag is not "deviceType" ? Generate error
                if (xmlNodeName(packageInfoXml,xmlDeviceTypeListChildNextSibling) != XML_IMPORT_EXPORT_NODE_DEVICETYPE)
                {
                  string sError = getCatStr("unDeviceTypeCreator","LOAD-INCORRECTXMLSTRUCTURE");
                  strreplace(sError,"TAG",xmlNodeName(packageInfoXml,xmlDeviceTypeListChildNextSibling));
      
                  fwException_raise(
                      dsExceptions,
                      "ERROR",
                      sError,
                      "LOAD-INCORRECTXMLSTRUCTURE"
                  ); 
                  return; 
                }
                
                // Read device type name
                string sDeviceTypeName;
                xmlGetElementAttribute(packageInfoXml,xmlDeviceTypeListChildNextSibling,XML_IMPORT_EXPORT_ATTR_NAME,sDeviceTypeName);
                if (sDeviceTypeName == "")
                {
                  string sError = getCatStr("unDeviceTypeCreator","LOAD-ATTRIBUTEEMPTYERROR");
                  strreplace(sError,"ATTR",XML_IMPORT_EXPORT_ATTR_NAME);
                  strreplace(sError,"NODE",xmlNodeName(packageInfoXml,xmlDeviceTypeListChildNextSibling));
    
                  fwException_raise(
                      dsExceptions,
                      "ERROR",
                      sError,
                      "LOAD-ATTRIBUTEEMPTYERROR"
                  ); 
                  return; 
                }  
                
                // Read device type exists (as a string "true"/"false")
                string sDeviceTypeExists;
                xmlGetElementAttribute(packageInfoXml,xmlDeviceTypeListChildNextSibling,XML_IMPORT_EXPORT_ATTR_EXISTS,sDeviceTypeExists);
                // Convert to bool
                bool bDeviceTypeExists;
                switch (sDeviceTypeExists)
                {
                  case "TRUE":
                  case "true":
                  {
                    bDeviceTypeExists = true;
                    break;                  
                  }
                  case "FALSE":
                  case "false":
                  {
                    bDeviceTypeExists = false;
                    break;                  
                  }
                  
                  default:
                  {
                    string sError = getCatStr("unDeviceTypeCreator","LOAD-INVALIDXMLBOOL");
                    strreplace(sError,"VALUE",sDeviceTypeExists);
                    strreplace(sError,"NODE",xmlNodeName(packageInfoXml,xmlDeviceTypeListChildNextSibling));
    
                    fwException_raise(
                        dsExceptions,
                        "ERROR",
                        sError,
                        "LOAD-INVALIDXMLBOOL"
                    ); 
                    return; 
                  }
                }
                dynAppend(localDeviceList,makeDynAnytype(sDeviceTypeName,bDeviceTypeExists));
                xmlDeviceTypeListChildNextSibling = xmlNextSibling(packageInfoXml, xmlDeviceTypeListChildNextSibling);
              }
              break;
            }
            // Widget type list
            case XML_IMPORT_EXPORT_NODE_WIDGETLIST:
            {
              // ---------------------------------------------------------------- //
              // ---------------------- Widget type list ------------------------ //
              // ---------------------------------------------------------------- //
              unsigned xmlWidgetTypeListChildNextSibling = xmlFirstChild(packageInfoXml,xmlListsFirstChildNextSibling);
              while (xmlWidgetTypeListChildNextSibling != -1)
              {
                // Child tag is not "widgetType" ? Generate error
                if (xmlNodeName(packageInfoXml,xmlWidgetTypeListChildNextSibling) != XML_IMPORT_EXPORT_NODE_WIDGETTYPE)
                {
                  string sError = getCatStr("unDeviceTypeCreator","LOAD-INCORRECTXMLSTRUCTURE");
                  strreplace(sError,"TAG",xmlNodeName(packageInfoXml,xmlWidgetTypeListChildNextSibling));
      
                  fwException_raise(
                      dsExceptions,
                      "ERROR",
                      sError,
                      "LOAD-INCORRECTXMLSTRUCTURE"
                  ); 
                  return; 
                }
                
                // Read widget type name
                string sWidgetTypeName;
                xmlGetElementAttribute(packageInfoXml,xmlWidgetTypeListChildNextSibling,XML_IMPORT_EXPORT_ATTR_NAME,sWidgetTypeName);
                if (sWidgetTypeName == "")
                {
                  string sError = getCatStr("unDeviceTypeCreator","LOAD-ATTRIBUTEEMPTYERROR");
                  strreplace(sError,"ATTR",XML_IMPORT_EXPORT_ATTR_NAME);
                  strreplace(sError,"NODE",xmlNodeName(packageInfoXml,xmlWidgetTypeListChildNextSibling));
    
                  fwException_raise(
                      dsExceptions,
                      "ERROR",
                      sError,
                      "LOAD-ATTRIBUTEEMPTYERROR"
                  ); 
                  return; 
                }  
								
								if (EMPTY_WIDGET_IDENTIFIER == sWidgetTypeName)
								{
									sWidgetTypeName = "";
								}
                
                // Read widget type exists (as a string "true"/"false")
                string sWidgetTypeExists;
                xmlGetElementAttribute(packageInfoXml,xmlWidgetTypeListChildNextSibling,XML_IMPORT_EXPORT_ATTR_EXISTS,sWidgetTypeExists);
                // Convert to bool
                bool bWidgetTypeExists;
                switch (sWidgetTypeExists)
                {
                  case "TRUE":
                  case "true":
                  {
                    bWidgetTypeExists = true;
                    break;                  
                  }
                  case "FALSE":
                  case "false":
                  {
                    bWidgetTypeExists = false;
                    break;                  
                  }
                  
                  default:
                  {
                    string sError = getCatStr("unDeviceTypeCreator","LOAD-INVALIDXMLBOOL");
                    strreplace(sError,"VALUE",sWidgetTypeExists);
                    strreplace(sError,"NODE",xmlNodeName(packageInfoXml,xmlWidgetTypeListChildNextSibling));
    
                    fwException_raise(
                        dsExceptions,
                        "ERROR",
                        sError,
                        "LOAD-INVALIDXMLBOOL"
                    ); 
                    return; 
                  }
                }
                dynAppend(localWidgetList,makeDynAnytype(sWidgetTypeName,bWidgetTypeExists));
                xmlWidgetTypeListChildNextSibling = xmlNextSibling(packageInfoXml, xmlWidgetTypeListChildNextSibling);
              }
              break;
            
              break;
            }
            // Cannot be something else. Raise error
            default:
            {
              string sError = getCatStr("unDeviceTypeCreator","LOAD-INCORRECTXMLSTRUCTURE");
              strreplace(sError,"TAG",xmlNodeName(packageInfoXml,xmlListsFirstChildNextSibling));
      
              fwException_raise(
                  dsExceptions,
                  "ERROR",
                  sError,
                  "LOAD-INCORRECTXMLSTRUCTURE"
              ); 
              return; 
            }
          }
          
          xmlListsFirstChildNextSibling = xmlNextSibling(packageInfoXml, xmlListsFirstChildNextSibling);
        }
        break;
      
      }
      case XML_IMPORT_EXPORT_NODE_MAPPING:
      {
        unsigned xmlMappingFirstChildNextSibling = xmlFirstChild(packageInfoXml,nextSibling);
        while (xmlMappingFirstChildNextSibling != -1)
        {
          switch (xmlNodeName(packageInfoXml,xmlMappingFirstChildNextSibling))
          {
          
            // ----------------------------------------------------------- //
            // ------------------ Front-end/device mapping --------------- //
            // ----------------------------------------------------------- //
            case XML_IMPORT_EXPORT_NODE_FRONTENDDEVICE:
            {
              unsigned xmlMappingFrontEndDeviceNextSibling = xmlFirstChild(packageInfoXml,xmlMappingFirstChildNextSibling);
              while (xmlMappingFrontEndDeviceNextSibling != -1)
              {
                if (xmlNodeName(packageInfoXml,xmlMappingFrontEndDeviceNextSibling) != XML_IMPORT_EXPORT_NODE_FRONTENDTYPE)
                {
                  string sError = getCatStr("unDeviceTypeCreator","LOAD-INCORRECTXMLSTRUCTURE");
                  strreplace(sError,"TAG",xmlNodeName(packageInfoXml,xmlMappingFrontEndDeviceNextSibling));
      
                  fwException_raise(
                      dsExceptions,
                      "ERROR",
                      sError,
                      "LOAD-INCORRECTXMLSTRUCTURE"
                  ); 
                  return; 
                }  
                
                string sFrontEndTypeName;
                xmlGetElementAttribute(packageInfoXml,xmlMappingFrontEndDeviceNextSibling,XML_IMPORT_EXPORT_ATTR_NAME,sFrontEndTypeName);
                                
                // Was this front-end in the initial list ?                
                if (dynContains(getDynString(localFrontEndList,LIST_NAME_INDEX),sFrontEndTypeName) < 1)
                {
                  string sError = getCatStr("unDeviceTypeCreator","LOAD-UNKNOWNFRONTENDTYPE");
                    strreplace(sError,"FRONTEND",sFrontEndTypeName);
    
                    fwException_raise(
                        dsExceptions,
                        "ERROR",
                        sError,
                        "LOAD-UNKNOWNFRONTENDTYPE"
                    ); 
                    return; 
                }
                
                // Mapping initialization for this front-end
                // There should not already be a key for this front-end, else it means the XML file is incorrect
                if (mappingHasKey(localFrontEndDeviceMapping,sFrontEndTypeName))
                {
                  string sError = getCatStr("unDeviceTypeCreator","LOAD-MULTIPLEFRONTENDDEFINITION");
                  strreplace(sError,"FRONTEND",sFrontEndTypeName);
      
                  fwException_raise(
                      dsExceptions,
                      "ERROR",
                      sError,
                      "LOAD-MULTIPLEFRONTENDDEFINITION"
                  ); 
                  return;                 
                }
                localFrontEndDeviceMapping[sFrontEndTypeName] = makeDynString();
                
                // Fill the mapping element with all the devices
                unsigned xmlMappingFrontEndDeviceNextDevice = xmlFirstChild(packageInfoXml,xmlMappingFrontEndDeviceNextSibling);
                while (xmlMappingFrontEndDeviceNextDevice != -1)
                {
                  if (xmlNodeName(packageInfoXml,xmlMappingFrontEndDeviceNextDevice) != XML_IMPORT_EXPORT_NODE_DEVICETYPE)
                  {
                    string sError = getCatStr("unDeviceTypeCreator","LOAD-INCORRECTXMLSTRUCTURE");
                    strreplace(sError,"TAG",xmlNodeName(packageInfoXml,xmlMappingFrontEndDeviceNextDevice));
      
                    fwException_raise(
                        dsExceptions,
                        "ERROR",
                        sError,
                        "LOAD-INCORRECTXMLSTRUCTURE"
                    ); 
                    return; 
                  }  
                  
                  string sDeviceTypeName;
                  xmlGetElementAttribute(packageInfoXml,xmlMappingFrontEndDeviceNextDevice,XML_IMPORT_EXPORT_ATTR_NAME,sDeviceTypeName);
                  
                  // Was this device in the initial list ?                
                  if (dynContains(getDynString(localDeviceList,LIST_NAME_INDEX),sDeviceTypeName) < 1)
                  {
                    string sError = getCatStr("unDeviceTypeCreator","LOAD-UNKNOWNDEVICETYPE");
                      strreplace(sError,"DEVICE",sDeviceTypeName);
    
                      fwException_raise(
                          dsExceptions,
                          "ERROR",
                          sError,
                          "LOAD-UNKNOWNDEVICETYPE"
                      ); 
                      return; 
                  }
                
                  dynAppend(localFrontEndDeviceMapping[sFrontEndTypeName],sDeviceTypeName);
                  xmlMappingFrontEndDeviceNextDevice = xmlNextSibling(packageInfoXml, xmlMappingFrontEndDeviceNextDevice);
                }
                  
                xmlMappingFrontEndDeviceNextSibling = xmlNextSibling(packageInfoXml, xmlMappingFrontEndDeviceNextSibling);
              }
              break;            
            }

            // ----------------------------------------------------------- //
            // ------------------- Device/widget mapping ----------------- //
            // ----------------------------------------------------------- //
            case XML_IMPORT_EXPORT_NODE_DEVICEWIDGET:
            {
              unsigned xmlMappingDeviceWidgetNextSibling = xmlFirstChild(packageInfoXml,xmlMappingFirstChildNextSibling);
              while (xmlMappingDeviceWidgetNextSibling != -1)
              {
                if (xmlNodeName(packageInfoXml,xmlMappingDeviceWidgetNextSibling) != XML_IMPORT_EXPORT_NODE_DEVICETYPE)
                {
                  string sError = getCatStr("unDeviceTypeCreator","LOAD-INCORRECTXMLSTRUCTURE");
                  strreplace(sError,"TAG",xmlNodeName(packageInfoXml,xmlMappingDeviceWidgetNextSibling));
      
                  fwException_raise(
                      dsExceptions,
                      "ERROR",
                      sError,
                      "LOAD-INCORRECTXMLSTRUCTURE"
                  ); 
                  return; 
                }  
                
                string sDeviceTypeName;
                xmlGetElementAttribute(packageInfoXml,xmlMappingDeviceWidgetNextSibling,XML_IMPORT_EXPORT_ATTR_NAME,sDeviceTypeName);
                                
                // Was this front-end in the initial list ?                
                if (dynContains(getDynString(localDeviceList,LIST_NAME_INDEX),sDeviceTypeName) < 1)
                {
                  string sError = getCatStr("unDeviceTypeCreator","LOAD-UNKNOWNFRONTENDTYPE");
                    strreplace(sError,"DEVICE",sDeviceTypeName);
    
                    fwException_raise(
                        dsExceptions,
                        "ERROR",
                        sError,
                        "LOAD-UNKNOWNFRONTENDTYPE"
                    ); 
                    return; 
                }
                
                // Mapping initialization for this front-end
                // There should not already be a key for this device, else it means the XML file is incorrect
                if (mappingHasKey(localDeviceWidgetMapping,sDeviceTypeName))
                {
                  string sError = getCatStr("unDeviceTypeCreator","LOAD-MULTIPLEDEVICEDEFINITION");
                  strreplace(sError,"DEVICE",sDeviceTypeName);
      
                  fwException_raise(
                      dsExceptions,
                      "ERROR",
                      sError,
                      "LOAD-MULTIPLEDEVICEDEFINITION"
                  ); 
                  return;                 
                }
                localDeviceWidgetMapping[sDeviceTypeName] = makeDynString();
                
                // Fill the mapping element with all the devices
                unsigned xmlMappingDeviceWidgetNextWidget = xmlFirstChild(packageInfoXml,xmlMappingDeviceWidgetNextSibling);
                while (xmlMappingDeviceWidgetNextWidget != -1)
                {
                  if (xmlNodeName(packageInfoXml,xmlMappingDeviceWidgetNextWidget) != XML_IMPORT_EXPORT_NODE_WIDGETTYPE)
                  {
                    string sError = getCatStr("unDeviceTypeCreator","LOAD-INCORRECTXMLSTRUCTURE");
                    strreplace(sError,"TAG",xmlNodeName(packageInfoXml,xmlMappingDeviceWidgetNextWidget));
      
                    fwException_raise(
                        dsExceptions,
                        "ERROR",
                        sError,
                        "LOAD-INCORRECTXMLSTRUCTURE"
                    ); 
                    return; 
                  }  
                  
                  string sWidgetTypeName;
                  xmlGetElementAttribute(packageInfoXml,xmlMappingDeviceWidgetNextWidget,XML_IMPORT_EXPORT_ATTR_NAME,sWidgetTypeName);
                  
									if (EMPTY_WIDGET_IDENTIFIER == sWidgetTypeName)
									{
										sWidgetTypeName = "";	
									}
                  // Was this device in the initial list ?                
                  if (dynContains(getDynString(localWidgetList,LIST_NAME_INDEX),sWidgetTypeName) < 1)
                  {
                    string sError = getCatStr("unDeviceTypeCreator","LOAD-UNKNOWNWIDGETTYPE");
                      strreplace(sError,"WIDGET",sWidgetTypeName);
    
                      fwException_raise(
                          dsExceptions,
                          "ERROR",
                          sError,
                          "LOAD-UNKNOWNWIDGETTYPE"
                      ); 
                      return; 
                  }
                
                  dynAppend(localDeviceWidgetMapping[sDeviceTypeName],sWidgetTypeName);
                  xmlMappingDeviceWidgetNextWidget = xmlNextSibling(packageInfoXml, xmlMappingDeviceWidgetNextWidget);
                }
                  
                xmlMappingDeviceWidgetNextSibling = xmlNextSibling(packageInfoXml, xmlMappingDeviceWidgetNextSibling);
              }
              break;            
            }
            // Cannot be something else. Raise error
            default:
            {
              string sError = getCatStr("unDeviceTypeCreator","LOAD-INCORRECTXMLSTRUCTURE");
              strreplace(sError,"TAG",xmlNodeName(packageInfoXml,xmlMappingFirstChildNextSibling));
      
              fwException_raise(
                  dsExceptions,
                  "ERROR",
                  sError,
                  "LOAD-INCORRECTXMLSTRUCTURE"
              ); 
              return; 
            }
          }
          
          xmlMappingFirstChildNextSibling = xmlNextSibling(packageInfoXml, xmlMappingFirstChildNextSibling);
        }
        break;
      }     
      default:
      { 
        string sError = getCatStr("unDeviceTypeCreator","LOAD-INCORRECTXMLSTRUCTURE");
        strreplace(sError,"TAG",xmlNodeName(packageInfoXml,nextSibling));
      
        fwException_raise(
            dsExceptions,
            "ERROR",
            sError,
            "LOAD-INCORRECTXMLSTRUCTURE"
        ); 
        return; 
      }
    }
      
    nextSibling = xmlNextSibling(packageInfoXml, nextSibling);
        
  }
  
  
  
  
  // Update the global variables	
  mappingClear(g_mFrontEndToDeviceMapping);
  mappingClear(g_mDeviceToWidgetMapping);
	dynClear(g_ddaFrontEndTypeList);
	dynClear(g_ddaDeviceTypeList);
	dynClear(g_ddaWidgetTypeList);
	
  g_mFrontEndToDeviceMapping = localFrontEndDeviceMapping;
	
  g_mDeviceToWidgetMapping = localDeviceWidgetMapping;  
	
	dynAppend(g_ddaFrontEndTypeList, localFrontEndList);
	
	dynAppend(g_ddaDeviceTypeList, localDeviceList);
	
	dynAppend(g_ddaWidgetTypeList, localWidgetList);
		
	g_sPackagePrefix = sPrefix;
	g_sPackageName = sPackageName;
	
}

//_unDeviceTypeCreator_getMappingForType
/**
  @par Description:
  Return the mapping of the asked type (front-end to device or device to widget). This allows a more generic program, the content of the variable is used regardless of it's meaning.
  
  @par Usage: 
  Internal 
	
	@param iTypeType input, indicates wether it is a front-end (CURRENT_TYPE_FRONT_END), a device (CURRENT_TYPE_DEVICE), a widget type (CURRENT_TYPE_WIDGET) or none (CURRENT_TYPE_NONE). The last two will return an empty mapping.
*/
mapping _unDeviceTypeCreator_getMappingForType(int iType)
{
  switch (iType)
  {
    case CURRENT_TYPE_FRONT_END:
    {
      return g_mFrontEndToDeviceMapping;
    }
    
    case CURRENT_TYPE_DEVICE: 
    {
      return g_mDeviceToWidgetMapping;
    }
    default:
    {
      break;
    }
  } 
  
  mapping m;
  return m;
}

//_unDeviceTypeCreator_getSuperTypes
/**
  @par Description:
  Returns all front-end to which a device is linked OR all devices to which a widget is linked.
  
  @par Usage: 
  Internal 
	
	@param iTypeType input, indicates wether it is a front-end (CURRENT_TYPE_FRONT_END), a device (CURRENT_TYPE_DEVICE), a widget type (CURRENT_TYPE_WIDGET) or none (CURRENT_TYPE_NONE). The last two will return an empty mapping.
*/
dyn_string _unDeviceTypeCreator_getSuperTypes(int iType, string sTypeName)
{
  
  dyn_string dsSuperTypesList;

  switch(iType)
  {
    case CURRENT_TYPE_DEVICE: 
    {
      for (int i = 1 ; i <= mappinglen(g_mFrontEndToDeviceMapping) ; i++)
      {
        if (dynContains(mappingGetValue(g_mFrontEndToDeviceMapping,i),sTypeName))
        {
          dynAppend(dsSuperTypesList,mappingGetKey(g_mFrontEndToDeviceMapping,i));
        }
      }  
      break;
    }
      
    case CURRENT_TYPE_WIDGET:
    {
      for (int i = 1 ; i <= mappinglen(g_mDeviceToWidgetMapping) ; i++)
      {
        if (dynContains(mappingGetValue(g_mDeviceToWidgetMapping,i),sTypeName))
        {
          dynAppend(dsSuperTypesList,mappingGetKey(g_mDeviceToWidgetMapping,i));
        }
      }  
      break;
    }
      
    default:
      break;
  
  }


  return dsSuperTypesList;

}

//_unDeviceTypeCreator_showProgress
/**
  @par Description:
  Display the current progress of the generation.
	If called from panel, it uses a progress bar. Otherwise it is printed as a debug in the console.
  
  @par Usage: 
  Internal 

	@param sText1    input, main text.
	@param sText2    input,	subtext one.
	@param sText3    input,	subtext two.
	@param iProgress input, the current progress (percentage).
*/
void _unDeviceTypeCreator_showProgress(string sText1, string sText2, string sText3, int iProgress)
{
	if (g_bUiActive)
	{
		showProgressBar(sText1, sText2, sText3, iProgress);
	}
	else
	{
    string sText = sText1 + "\n";
    sText += ("" != sText2)? "\t" + sText2 + "\n" : "";
    sText += ("" != sText3)? "\t" + sText3 + "\n" : "";
    sText += "Progress: " + iProgress + "%";
    
		DebugN(sText);
	}

}

//_unDeviceTypeCreator_fileExists
/**
  @par Description:
  Check is a file already exists, so it is not generated again.
  
  @par Usage: 
  Internal 

	@param sFileName input, the file to check.
*/
bool _unDeviceTypeCreator_fileExists(string sFileName)
{
  return dynContains(dsLoadedPackageFiles, sFileName) > 0;
}



//@}
