/**@name LIBRARY: unLHCLoggingDBSettings.ctl

@author: Josef Hofer (EN-ICE-SCD)

Creation Date: 04/09/2014

Purpose:
This library contains settings related functions for the unLHCLoggingDB component.

Usage:
Library of the unLHCLoggingDB component used only by unLHCLoggingDB.

PVSS manager usage: CTRL, UI

Constraints:
	. PVSS version: 3.11
	. operating system: Linux.

 Copyright CERN

*/

//@{

public const string unLHCLoggingDBSettings_sDefaultConfigFileGeneratorFunction = "unLHCLoggingDBUserLib_getDeviceInfo_UNICOS";

// Datapoint that holds the name of the active configuration datapoint
private const string sSettingsDP = "unLHCLoggingDB_Settings";

// Datapoints that hold database settings
private const string sDbConfigDpType = "_UnLHCLoggingDB_Database";
private const string sDbConfigDpPrefix = "unLHCLoggingDB_Database_";


/**
* Get name of settings Dp
*
* @param sDbName     input,  name of the database
* @param sDbUser     input,  name of the database
*
* @return            output, name of the datapoint
*/
public string unLHCLoggingDBSettings_getDbConfigDpName(string sDbName, string sDbUser)
{
    return (getSystemName() + sDbConfigDpPrefix + sDbName + "_" + sDbUser);
}


/**
* Get name of settings dpe
*
* @param sDpeName     output, Name of the dpe that holds the settings
*
* @return             output, true: success, false: error
*/
public bool unLHCLoggingDBSettings_getSettingsDpeName(string &sDpeName)
{
    sDpeName = getSystemName() + sSettingsDP + ".activeDatabase";
    return dpExists(sDpeName);
}


/**
* Get name of datapoint, that holds database settings
*
* @param sDpName      output, Name of the database configuration datapoint
*
* @return             output, true: success, false: error
*/
public bool unLHCLoggingDBSettings_getActiveDbConfigDp(string &sDpName)
{
    string sSettingsDpe;
    if (unLHCLoggingDBSettings_getSettingsDpeName(sSettingsDpe) == true) {
        dpGet(sSettingsDpe, sDpName);
        return true;
    }
    return false;
}


/**
* Set name of datapoint, that holds database settings
*
* @param sDpName      input, Name of the database configuration datapoint
*
* @return             output, true: success, false: error
*/
public bool unLHCLoggingDBSettings_setActiveDbConfigDp(string sDpName)
{
    string sSettingsDpe;
    if (unLHCLoggingDBSettings_getSettingsDpeName(sSettingsDpe) == true) {
        dpSetWait(sSettingsDpe, sDpName);
        return true;
    }
    return false;
}


/**
* Get data from a specific database configuration datapoint
*
* @param sDbConfigDp    input, Name of the database configuration datapoint
* @param sDbName        output, database name
* @param sDbUser        output, database user
* @param sDbPass        output, database password (encrypted)
*
* @return               output, true: success, false: error
*/
public bool unLHCLoggingDBSettings_getDbConfig(string sDbConfigDp, string &sDbName, string &sDbUser, string &sDbPass)
{
    if (dpExists(sDbConfigDp)) {
        dpGet(sDbConfigDp + ".dbName", sDbName,
              sDbConfigDp + ".dbUser", sDbUser,
              sDbConfigDp + ".dbPass", sDbPass);
        return true;
    }
    return false;
}


/**
* Set data of a specific database configuration datapoint
*
* @param sDbConfigDp    input, Name of the database configuration datapoint
* @param sDbName        input, database name
* @param sDbUser        input, database user
* @param sDbPass        input, database password (encrypted)
*
* @return               output, true: success, false: error
*/
private bool unLHCLoggingDBSettings_setDbConfig(string sDbConfigDp, string sDbName, string sDbUser, string sDbPass)
{
    if (dpExists(sDbConfigDp)) {
        dpSetWait(sDbConfigDp + ".dbName", sDbName,
                  sDbConfigDp + ".dbUser", sDbUser,
                  sDbConfigDp + ".dbPass", sDbPass);
        return true;
    }
    return false;
}


/**
* Get database settings of active database
*
* @param sDbName        output, database name
* @param sDbUser        output, database user
* @param sDbPass        output, database password (encrypted)
*
* @return               output, true: success, false: error
*/
public bool unLHCLoggingDBSettings_getActiveDbConfig(string &sDbName, string &sDbUser, string &sDbPass)
{
    string sDbConfigDp;

    if (unLHCLoggingDBSettings_getActiveDbConfigDp(sDbConfigDp) == false) {
        return false;
    }

    return unLHCLoggingDBSettings_getDbConfig(sDbConfigDp, sDbName, sDbUser, sDbPass);
}


/**
* Get database settings of all database configuration datapoints
*
* @param dsDbConfigDps    output, database configuration datapoint names
* @param dsDbNames        output, database names
* @param dsDbUsers        output, database users
* @param dsDbPasswords    output, database passwords (encrypted)
* @param dbDpEditable     output, true=dbConfigDp is editable; false dbConfigDp is read only
*
* @return                 output, true: success, false: error
*/
public bool unLHCLoggingDBSettings_getAllDbConfigDps(dyn_string &dsDbConfigDps,
        dyn_string &dsDbNames,
        dyn_string &dsDbUsers,
        dyn_string &dsDbPasswords,
        dyn_bool &dbDpEditable)
{
    bool bSuccess = true;
    string sSystemName = getSystemName();

    dynClear(dsDbConfigDps);
    dynClear(dsDbNames);
    dynClear(dsDbUsers);
    dynClear(dsDbPasswords);
    dynClear(dbDpEditable);

    dsDbConfigDps = dpNames("*", "_UnLHCLoggingDB_Database");

    int iDsDbConfigDpsLen = dynlen(dsDbConfigDps);
    for (int i = 1; i <= iDsDbConfigDpsLen; i++) {
        // Only dbConfigDps of current system are editable
        dbDpEditable[i] = patternMatch(sSystemName + "*", dsDbConfigDps[i]);
        if (unLHCLoggingDBSettings_getDbConfig(dsDbConfigDps[i], dsDbNames[i], dsDbUsers[i], dsDbPasswords[i]) == false) {
            bSuccess = false;
        }
    }
    return bSuccess;
}


/**
* Set database settings of all database configuration datapoints
*
* @param dsDbConfigDps    input, database configuration datapoint names
* @param dsDbNames        input, database names
* @param dsDbUsers        input, database users
* @param dsDbPasswords    input, database passwords (encrypted)
*
* @return                 output, true: success, false: error
*/
public bool unLHCLoggingDBSettings_setAllDbConfigDps(const dyn_string &dsDbConfigDps, const dyn_string &dsDbNames, const dyn_string &dsDbUsers, const dyn_string &dsDbPasswords)
{
    bool bSuccess = true;
    string sSystemName = getSystemName();

    dyn_string dsDbConfigDpsOld = dpNames("*", "_UnLHCLoggingDB_Database");

    // Compare lists of old dbConfigDps with new ones and delete DPs that are not in the new list (because they have been deleted by the user)
    int iDsDbConfigDpsOldLen = dynlen(dsDbConfigDpsOld);
    for (int i = 1; i <= iDsDbConfigDpsOldLen; i++) {
        if (dynContains(dsDbConfigDps, dsDbConfigDpsOld[i]) == 0) {
            if (patternMatch(sSystemName + "*", dsDbConfigDpsOld[i]) == true) {
                // Only datapoints of this system can be deleted
                dpDelete(dsDbConfigDpsOld[i]);
                dyn_errClass err = getLastError();
                if (dynlen(err) > 0) {
                    bSucces = false;
                }
            }
        }
    }

    // Compare lists of new dbConfigDps with old ones and create DPs that are only in the new list (because they have been added by the user)
    int iDsDbConfigDpsLen = dynlen(dsDbConfigDps);
    for (int i = 1; i <= iDsDbConfigDpsLen; i++) {
        // Check if new element already exists
        int iIndex = dynContains(dsDbConfigDpsOld, dsDbConfigDps[i]);
        if (iIndex > 0) { // it exists --> update it
            // Update values of datapoint
            if (patternMatch(sSystemName + "*", dsDbConfigDps[i]) == true) {
                // Only datapoints of this system can be changed
                if (!unLHCLoggingDBSettings_setDbConfig(dsDbConfigDps[i], dsDbNames[i], dsDbUsers[i], dsDbPasswords[i])) {
                    bSuccess = false;
                }
            }
        } else if (iIndex == 0) { // it does not exist -- create dp
            // Remove system name
            int iPos = strpos(dsDbConfigDps[i], ":");
            string sDpName = iPos >= 0 ? substr(dsDbConfigDps[i], iPos + 1) : dsDbConfigDps[i] ;

            dpCreate(sDpName, sDbConfigDpType);
            dyn_errClass err = getLastError();
            if (dynlen(err) > 0) {
                bSucces = false;
            } else {
                if (!unLHCLoggingDBSettings_setDbConfig(dsDbConfigDps[i], dsDbNames[i], dsDbUsers[i], dsDbPasswords[i])) {
                    bSuccess = false;
                }
            }
        }
    }

    return bSuccess;
}


/**
* Get db user from the archive configuration of the current system
*
* @return   mySystemDbSchema     output, Schema name as string
*/
public string unLHCLoggingDBSettings_getMySystemDbSchema()
{
    string sSchema;
    dpGet("_RDBArchive.db.user", sSchema);
    return sSchema;
}


/**
* Read mode of time stamps from datapoint
*
* @return   bUTC       output, 0=localtime 1=UTC
*/
public bool unLHCLoggingDBSettings_getTimeStampMode()
{
    bool bUTC;
    dpGet(getSystemName() + sSettingsDP + ".timeAsUTC", bUTC);
    return bUTC;
}


/**
* Write mode of time stamps to datapoint
*
* @param    bUTC       input, 0=localtime 1=UTC
* @return              output, true: success, false: error
*/
public bool unLHCLoggingDBSettings_setTimeStampMode(bool bUTC)
{
    dpSetWait(getSystemName() + sSettingsDP + ".timeAsUTC", bUTC);

    dyn_errClass err = getLastError();
    return (dynlen(err) == 0 ? true : false);
}

public string unLHCLoggingDBSettings_getConfigFileGeneratorFunction()
{
    string sFunction;
    dpGet(getSystemName() + sSettingsDP + ".function", sFunction);
    return sFunction;
}

public bool unLHCLoggingDBSettings_setConfigFileGeneratorFunction(string sFunction)
{
    dpSetWait(getSystemName() + sSettingsDP + ".function", sFunction);

    dyn_errClass err = getLastError();
    return (dynlen(err) == 0 ? true : false);
}

